"""Classes for handling models (sets of coordinates) as well as
   groups of models.
"""

import struct
import itertools


class Sphere(object):
    """Coordinates of part of the model represented by a sphere.

       See :meth:`Model.get_spheres` for more details.

       :param asym_unit: The asymmetric unit that this sphere represents
       :type asym_unit: :class:`ihm.AsymUnit`
       :param tuple seq_id_range: The range of residues represented by this
              sphere (as a two-element tuple)
       :param float x: x coordinate of the center of the sphere
       :param float y: y coordinate of the center of the sphere
       :param float z: z coordinate of the center of the sphere
       :param float radius: radius of the sphere
       :param float rmsf: root-mean-square fluctuation of the coordinates
    """

    # Reduce memory usage
    __slots__ = ['asym_unit', 'seq_id_range', 'x', 'y', 'z', 'radius', 'rmsf']

    def __init__(self, asym_unit, seq_id_range, x, y, z, radius, rmsf=None):
        self.asym_unit = asym_unit
        self.seq_id_range = seq_id_range
        self.x, self.y, self.z = x, y, z
        self.radius, self.rmsf = radius, rmsf


class Atom(object):
    """Coordinates of part of the model represented by an atom.

       See :meth:`Model.get_atoms` for more details. Note that this class
       is used only to represent the coordinates of an atom. To access
       atom-specific properties of the model, see the :class:`ihm.Atom` class.

       :param asym_unit: The asymmetric unit that this atom represents
       :type asym_unit: :class:`ihm.AsymUnit`
       :param int seq_id: The residue index represented by this atom
              (can be None for HETATM sites)
       :param str atom_id: The name of the atom in the residue
       :param str type_symbol: Element name
       :param float x: x coordinate of the atom
       :param float y: y coordinate of the atom
       :param float z: z coordinate of the atom
       :param bool het: True for HETATM sites, False (default) for ATOM
       :param float biso: Temperature factor or equivalent (if applicable)
       :param float occupancy: Fraction of the atom type present
              (if applicable)
    """

    # Reduce memory usage
    __slots__ = ['asym_unit', 'seq_id', 'atom_id', 'type_symbol',
                 'x', 'y', 'z', 'het', 'biso', 'occupancy']

    def __init__(self, asym_unit, seq_id, atom_id, type_symbol, x, y, z,
                 het=False, biso=None, occupancy=None):
        self.asym_unit = asym_unit
        self.seq_id, self.atom_id = seq_id, atom_id
        self.type_symbol = type_symbol
        self.x, self.y, self.z = x, y, z
        self.het, self.biso = het, biso
        self.occupancy = occupancy


class Model(object):
    """A single set of coordinates (conformation).

       Models are added to the system by placing them inside
       :class:`ModelGroup` objects, which in turn are placed inside
       :class:`State` objects, which are grouped in
       :class:`StateGroup` objects, which are finally added to the system
       via :attr:`ihm.System.state_groups`.

       :param assembly: The parts of the system that were modeled.
       :type assembly: :class:`~ihm.Assembly`
       :param protocol: Description of how the modeling was done.
       :type protocol: :class:`~ihm.protocol.Protocol`
       :param representation: Level of detail at which the system
              was represented.
       :type representation: :class:`~ihm.representation.Representation`
       :param str name: Descriptive name for this model.
    """
    def __init__(self, assembly, protocol, representation, name=None):
        self.assembly, self.protocol = assembly, protocol
        self.representation, self.name = representation, name
        self._atoms = []
        self._spheres = []

    def get_spheres(self):
        """Yield :class:`Sphere` objects that represent this model.

           The default implementation simply iterates over an internal
           list of spheres, but this is not very memory-efficient, particularly
           if the spheres are already stored somewhere else, e.g. in the
           software's own data structures. It is recommended to subclass
           and provide a more efficient implementation. For example, the
           `modeling of Nup133 <https://github.com/integrativemodeling/nup133/>`_
           uses a `custom subclass <https://github.com/integrativemodeling/nup133/blob/main/outputs_foxs_ensemble_new/pdb-dev/pdb.py>`_
           to pass `BioPython <https://biopython.org/>`_ objects through
           to python-ihm.

           Note that the set of spheres should match the model's
           :class:`~ihm.representation.Representation`. This is not currently
           enforced.
        """  # noqa: E501
        for s in self._spheres:
            yield s

    def add_sphere(self, sphere):
        """Add to the model's set of :class:`Sphere` objects.

           See :meth:`get_spheres` for more details.
        """
        self._spheres.append(sphere)

    def get_atoms(self):
        """Yield :class:`Atom` objects that represent this model.

           See :meth:`get_spheres` for more details.
        """
        for a in self._atoms:
            yield a

    def add_atom(self, atom):
        """Add to the model's set of :class:`Atom` objects.

           See :meth:`get_spheres` for more details.
        """
        self._atoms.append(atom)


class ModelGroup(list):
    """A set of related models. See :class:`Model`. It is implemented as
       a simple list of the models.

       These objects are typically stored in a :class:`State`,
       :class:`Ensemble`, or :class:`OrderedProcess`.

       :param elements: Initial set of models in the group.
       :param str name: Descriptive name for the group.
    """
    def __init__(self, elements=(), name=None):
        self.name = name
        super(ModelGroup, self).__init__(elements)


class State(list):
    """A set of model groups that constitute a single state of the system.
       It is implemented as a simple list of the model groups.
       See :class:`StateGroup`.

       :param elements: The initial set of :class:`ModelGroup` objects in
              this state.
    """
    def __init__(self, elements=(), type=None, name=None, details=None,
                 experiment_type=None, population_fraction=None):
        self.type, self.name, self.details = type, name, details
        self.experiment_type = experiment_type
        self.population_fraction = population_fraction
        super(State, self).__init__(elements)


class StateGroup(list):
    """A set of related states. See :class:`State` and
       :attr:`ihm.System.state_groups`. It is implemented as a simple
       list of the states.

       :param elements: Initial set of states in the group.
    """
    def __init__(self, elements=()):
        super(StateGroup, self).__init__(elements)


class Ensemble(object):
    """Details about a model cluster or ensemble.
       See :attr:`ihm.System.ensembles`.

       :param model_group: The set of models in this ensemble.
       :type model_group: :class:`ModelGroup`
       :param int num_models: The total number of models in this ensemble. This
              may be more than the number of models in `model_group`, for
              example if only representative or top-scoring models
              are deposited.
       :param post_process: The final analysis step that generated this
              ensemble.
       :type post_process: :class:`ihm.analysis.Step`
       :param str clustering_method: The method used to obtain the ensemble,
              if applicable.
       :param str clustering_feature: The feature used for clustering
              the models, if applicable.
       :param str name: A descriptive name for this ensemble.
       :param float precision: The precision of the entire ensemble.
       :param file: A reference to an external file containing coordinates
              for the entire ensemble, for example as a DCD file
              (see :class:`DCDWriter`). See also :attr:`subsamples`.
       :type file: :class:`ihm.location.OutputFileLocation`
       :param str details: Additional text describing this ensemble
    """
    def __init__(self, model_group, num_models, post_process=None,
                 clustering_method=None, clustering_feature=None, name=None,
                 precision=None, file=None, details=None):
        self.model_group, self.num_models = model_group, num_models
        self.post_process = post_process
        self.clustering_method = clustering_method
        self.clustering_feature = clustering_feature
        self.name, self.precision, self.file = name, precision, file
        self.details = details

        #: All localization densities for this ensemble, as
        #: :class:`LocalizationDensity` objects
        self.densities = []

        #: All subsamples that make up this ensemble (if applicable),
        #: as :class:`Subsample` objects
        self.subsamples = []

    num_models_deposited = property(lambda self: len(self.model_group),
                                    doc="Number of models in this ensemble "
                                        "that are in the mmCIF file")


class OrderedProcess(object):
    """Details about a process that orders two or more model groups.

       A process is represented as a directed graph, where the nodes
       are :class:`ModelGroup` objects and the edges represent transitions.

       These objects are generally added to
       :attr:`ihm.System.ordered_processes`.

       :param str ordered_by: Text that explains how the ordering is done,
              such as "time steps".
       :param str description: Text that describes this process.
    """
    def __init__(self, ordered_by, description=None):
        self.ordered_by, self.description = ordered_by, description
        #: All steps in this process, as a simple list of
        #: :class:`ProcessStep` objects
        self.steps = []


class ProcessStep(list):
    """A single step in an :class:`OrderedProcess`.

       This is implemented as a simple list of :class:`ProcessEdge` objects,
       each of which orders two :class:`ModelGroup` objects. (To order more
       than two groups, for example to represent a branched reaction step
       that generates two products, simply add multiple edges to the step.)

       :param sequence elements: Initial set of :class:`ProcessEdge` objects.
       :param str description: Text that describes this step.
    """
    def __init__(self, elements=(), description=None):
        self.description = description
        super(ProcessStep, self).__init__(elements)


class ProcessEdge(object):
    """A single directed edge in the graph for a :class:`OrderedProcess`,
       representing the transition from one :class:`ModelGroup` to another.
       These objects are added to :class:`ProcessStep` objects.

       :param group_begin: The set of models at the origin of the edge.
       :type group_begin: :class:`ModelGroup`
       :param group_end: The set of models at the end of the edge.
       :type group_end: :class:`ModelGroup`
       :param str description: Text that describes this edge.
    """
    def __init__(self, group_begin, group_end, description=None):
        self.group_begin, self.group_end = group_begin, group_end
        self.description = description


class LocalizationDensity(object):
    """Localization density of part of the system, over all models
       in an ensemble.

       See :attr:`Ensemble.densities`.

       :param file: A reference to an external file containing the density,
              for example as an MRC file.
       :type file: :class:`ihm.location.OutputFileLocation`
       :param asym_unit: The asymmetric unit (or part of one) that
              this density represents.
       :type asym_unit: :class:`~ihm.AsymUnit` or :class:`~ihm.AsymUnitRange`
    """
    def __init__(self, file, asym_unit):
        self.file, self.asym_unit = file, asym_unit


class Subsample(object):
    """Base class for a subsample within an ensemble.

       In some cases the models that make up an :class:`Ensemble` may be
       partitioned into subsamples, for example to determine if the
       sampling was exhaustive
       (see `Viswanath et al. 2017 <https://www.ncbi.nlm.nih.gov/pmc/articles/pmid/29211988/>`_).
       This base class can be used to describe the set of models in the
       subsample, for example by pointing to an externally-deposited
       set of conformations.

       Usually a derived class (:class:`RandomSubsample` or
       :class:`IndependentSubsample`) is used instead of this class.
       Instances are stored in :attr:`Ensemble.subsamples`. All of the
       subsamples in a given ensemble must be of the same type.

       :param str name: A descriptive name for this sample
       :param int num_models: The total number of models in this sample
       :param model_group: The set of models in this sample, if applicable.
       :type model_group: :class:`ModelGroup`
       :param file: A reference to an external file containing coordinates
              for the entire sample, for example as a DCD file
              (see :class:`DCDWriter`).
       :type file: :class:`ihm.location.OutputFileLocation`
    """  # noqa: E501

    sub_sampling_type = 'other'

    def __init__(self, name, num_models, model_group=None, file=None):
        self.name, self.num_models = name, num_models
        self.model_group, self.file = model_group, file

    num_models_deposited = property(
        lambda self: len(self.model_group) if self.model_group else 0,
        doc="Number of models in this subsample that are in the mmCIF file")


class RandomSubsample(Subsample):
    """A subsample generated by picking a random subset of the models that
       make up the entire ensemble. See :class:`Subsample`.
    """
    sub_sampling_type = 'random'


class IndependentSubsample(Subsample):
    """A subsample generated in the same fashion as other subsamples
       but by an independent simulation. See :class:`Subsample`.
    """
    sub_sampling_type = 'independent'


class DCDWriter(object):
    """Utility class to write model coordinates to a binary DCD file.

       See :class:`Ensemble` and :class:`Model`. Since mmCIF is a text-based
       format, it is not efficient to store entire ensembles in this format.
       Instead, representative models should be deposited as mmCIF and
       the :class:`Ensemble` then linked to an external file containing
       only model coordinates. One such format is CHARMM/NAMD's DCD, which
       is written out by this class. The DCD files simply contain the xyz
       coordinates of all :class:`Atom` and :class:`Sphere` objects in each
       :class:`Model`. (Note that no other data is stored, such as sphere
       radii or restraint parameters.)

       :param file fh: The filelike object to write the coordinates to. This
              should be open in binary mode and should be a seekable object.
    """
    def __init__(self, fh):
        self.fh = fh
        self.nframes = 0

    def add_model(self, model):
        """Add the coordinates for the given :class:`Model` to the file as
           a new frame. All models in the file should have the same number of
           atoms and/or spheres, in the same order.

           :param model: Model with coordinates to write to the file.
           :type model: :class:`Model`
        """
        x = []
        y = []
        z = []
        for a in itertools.chain(model.get_atoms(), model.get_spheres()):
            x.append(a.x)
            y.append(a.y)
            z.append(a.z)
        self._write_frame(x, y, z)

    def _write_frame(self, x, y, z):
        self.nframes += 1
        if self.nframes == 1:
            self.ncoord = len(x)
            remarks = [
                b'Produced by python-ihm, https://github.com/ihmwg/python-ihm',
                b'This file is designed to be used in combination with an '
                b'mmCIF file',
                b'See PDB-Dev at https://pdb-dev.wwpdb.org/ for more details']
            self._write_header(self.ncoord, remarks)
        else:
            if len(x) != self.ncoord:
                raise ValueError(
                    "Frame size mismatch - frames contain %d "
                    "coordinates but attempting to write a frame "
                    "containing %d coordinates" % (self.ncoord, len(x)))
            # Update number of frames
            self.fh.seek(self._pos_nframes)
            self.fh.write(struct.pack('i', self.nframes))
            self.fh.seek(0, 2)  # Move back to end of file

        # Write coordinates
        frame_size = struct.pack('i', struct.calcsize("%df" % self.ncoord))
        for coord in x, y, z:
            self.fh.write(frame_size)
            self.fh.write(struct.pack("%df" % self.ncoord, *coord))
            self.fh.write(frame_size)

    def _write_header(self, natoms, remarks):
        self.fh.write(struct.pack('i', 84) + b'CORD')
        self._pos_nframes = self.fh.tell()
        self.fh.write(struct.pack('i', self.nframes))
        self.fh.write(struct.pack('i', 0))  # istart
        self.fh.write(struct.pack('i', 0))  # nsavc
        self.fh.write(struct.pack('5i', 0, 0, 0, 0, 0))
        self.fh.write(struct.pack('i', 0))  # number of fixed atoms
        self.fh.write(struct.pack('d', 0.))  # delta
        self.fh.write(struct.pack('10i', 0, 0, 0, 0, 0, 0, 0, 0, 0, 84))
        remark_size = struct.calcsize('i') + 80 * len(remarks)
        self.fh.write(struct.pack('i', remark_size))
        self.fh.write(struct.pack('i', len(remarks)))
        for r in remarks:
            self.fh.write(r.ljust(80)[:80])
        self.fh.write(struct.pack('i', remark_size))
        self.fh.write(struct.pack('i', struct.calcsize('i')))
        self.fh.write(struct.pack('i', natoms))  # total number of atoms
        self.fh.write(struct.pack('i', struct.calcsize('i')))



"""Classes for tracking external data used by mmCIF models.
"""

import os


class Location(object):
    """Identifies the location where a resource can be found.

       Do not use this class itself, but one of its subclasses.
       Typically the resource may be found in a file (either on the local
       disk or at a DOI) - for this use one of the subclasses of
       :class:`FileLocation`. Alternatively the resource may be found in
       an experiment-specific database such as PDB or EMDB - for this use
       :class:`DatabaseLocation` or one of its subclasses. A Location may
       be passed to

         - a :class:`~ihm.dataset.Dataset` to point to where an
           experimental dataset may be found;
         - an :class:`~ihm.model.Ensemble` to point to coordinates for an
           entire ensemble, for example as a DCD file;
         - a :class:`LocalizationDensity` to point to an external localization
           density, for example in MRC format;
         - :data:`ihm.System.locations` to point to other files relating
           to the modeling in general, such as a modeling control script
           (:class:`WorkflowFileLocation`) or a command script for a
           visualization package such as ChimeraX
           (:class:`VisualizationFileLocation`);
         - a :class:`ihm.protocol.Step` or :class:`ihm.analysis.Step` to
           describe an individual modeling step;
         - or a :class:`~ihm.startmodel.StartingModel` to describe how a
           starting model was constructed.

       :param str details: Additional details about the dataset, if known.

    """

    # 'details' can differ without affecting dataset equality
    _eq_keys = []
    _allow_duplicates = False

    def __init__(self, details=None):
        self.details = details

    # Locations compare equal iff they are the same class, have the
    # same attributes, and allow_duplicates=False
    def _eq_vals(self):
        if self._allow_duplicates:
            return id(self)
        else:
            return tuple([self.__class__]
                         + [getattr(self, x) for x in self._eq_keys])

    def __eq__(self, other):
        # We can never be equal to None
        return other is not None and self._eq_vals() == other._eq_vals()

    def __hash__(self):
        return hash(self._eq_vals())


class DatabaseLocation(Location):
    """A dataset stored in an official database (PDB, EMDB, PRIDE, etc.).

       :param str db_name: The name of the database.
       :param str db_code: The accession code inside the database.
       :param str version: The version of the dataset in the database.
       :param str details: Additional details about the dataset, if known.
    """

    _eq_keys = Location._eq_keys + ['db_name', 'access_code', 'version']

    def __init__(self, db_name, db_code, version=None, details=None):
        super(DatabaseLocation, self).__init__(details)
        self.db_name = db_name
        self.access_code = db_code
        self.version = version


class EMDBLocation(DatabaseLocation):
    """Something stored in the EMDB database.
       See :class:`DatabaseLocation` for a description of the parameters
       and :class:`Location` for discussion of the usage of these objects."""
    _db_name = 'EMDB'

    def __init__(self, db_code, version=None, details=None):
        super(EMDBLocation, self).__init__(self._db_name, db_code,
                                           version, details)


class PDBLocation(DatabaseLocation):
    """Something stored in the PDB database.
       See :class:`DatabaseLocation` for a description of the parameters
       and :class:`Location` for discussion of the usage of these objects."""
    _db_name = 'PDB'

    def __init__(self, db_code, version=None, details=None):
        super(PDBLocation, self).__init__(self._db_name, db_code,
                                          version, details)


class PDBDevLocation(DatabaseLocation):
    """Something stored in the PDB-Dev database.
       See :class:`DatabaseLocation` for a description of the parameters
       and :class:`Location` for discussion of the usage of these objects."""
    _db_name = 'PDB-Dev'

    def __init__(self, db_code, version=None, details=None):
        super(PDBDevLocation, self).__init__(self._db_name, db_code,
                                             version, details)


class BMRBLocation(DatabaseLocation):
    """Something stored in the BMRB database.
       See :class:`DatabaseLocation` for a description of the parameters
       and :class:`Location` for discussion of the usage of these objects."""
    _db_name = 'BMRB'

    def __init__(self, db_code, version=None, details=None):
        super(BMRBLocation, self).__init__(self._db_name, db_code,
                                           version, details)


class MassIVELocation(DatabaseLocation):
    """Something stored in the MassIVE database.
       See :class:`DatabaseLocation` for a description of the parameters
       and :class:`Location` for discussion of the usage of these objects."""
    _db_name = 'MASSIVE'

    def __init__(self, db_code, version=None, details=None):
        super(MassIVELocation, self).__init__(self._db_name, db_code, version,
                                              details)


class EMPIARLocation(DatabaseLocation):
    """Something stored in the EMPIAR database.
       See :class:`DatabaseLocation` for a description of the parameters
       and :class:`Location` for discussion of the usage of these objects."""
    _db_name = 'EMPIAR'

    def __init__(self, db_code, version=None, details=None):
        super(EMPIARLocation, self).__init__(self._db_name, db_code, version,
                                             details)


class SASBDBLocation(DatabaseLocation):
    """Something stored in the SASBDB database.
       See :class:`DatabaseLocation` for a description of the parameters
       and :class:`Location` for discussion of the usage of these objects."""
    _db_name = 'SASBDB'

    def __init__(self, db_code, version=None, details=None):
        super(SASBDBLocation, self).__init__(self._db_name, db_code, version,
                                             details)


class PRIDELocation(DatabaseLocation):
    """Something stored in the PRIDE database.
       See :class:`DatabaseLocation` for a description of the parameters
       and :class:`Location` for discussion of the usage of these objects."""
    _db_name = 'PRIDE'

    def __init__(self, db_code, version=None, details=None):
        super(PRIDELocation, self).__init__(self._db_name, db_code, version,
                                            details)


class BioGRIDLocation(DatabaseLocation):
    """Something stored in the BioGRID database.
       See :class:`DatabaseLocation` for a description of the parameters
       and :class:`Location` for discussion of the usage of these objects."""
    _db_name = 'BioGRID'

    def __init__(self, db_code, version=None, details=None):
        super(BioGRIDLocation, self).__init__(self._db_name, db_code, version,
                                              details)


class ProXLLocation(DatabaseLocation):
    """Something stored in the ProXL database.
       See :class:`DatabaseLocation` for a description of the parameters
       and :class:`Location` for discussion of the usage of these objects."""
    _db_name = 'ProXL'

    def __init__(self, db_code, version=None, details=None):
        super(ProXLLocation, self).__init__(self._db_name, db_code, version,
                                            details)


class FileLocation(Location):
    """Base class for an individual file or directory stored externally.

       :param str path: the location of the file or directory (this can
              be `None` if `repo` is set, to refer to the entire repository)
       :param repo: object that describes the repository
              containing the file, or `None` if it is stored on the local disk
       :type repo: :class:`Repository`
       :param str details: optional description of the file
    """

    _eq_keys = Location._eq_keys + ['repo', 'path', 'content_type']

    content_type = None

    def __init__(self, path, repo=None, details=None):
        super(FileLocation, self).__init__(details)
        self.repo = repo
        if repo:
            self.path = path
            # Cannot determine file size if non-local
            self.file_size = None
        else:
            if not os.path.exists(path):
                raise ValueError("%s does not exist" % path)
            self.file_size = os.stat(path).st_size
            # Store absolute path in case the working directory changes later
            self.path = os.path.abspath(path)


class InputFileLocation(FileLocation):
    """An externally stored file used as input.
       See :class:`FileLocation` for a description of the parameters
       and :class:`Location` for discussion of the usage of these objects.

       For example, any :class:`~ihm.dataset.Dataset` that isn't stored in
       a domain-specific database would use this class."""
    content_type = 'Input data or restraints'


class OutputFileLocation(FileLocation):
    """An externally stored file used for output.
       See :class:`FileLocation` for a description of the parameters
       and :class:`Location` for discussion of the usage of these objects.

       For example, this can be used to point to an externally-stored
       :class:`model ensemble <ihm.model.Ensemble>` or a
       :class:`localization density <ihm.model.LocalizationDensity>`.
    """
    content_type = "Modeling or post-processing output"


class WorkflowFileLocation(FileLocation):
    """An externally stored file that controls the workflow (e.g. a script).
       See :class:`FileLocation` for a description of the parameters
       and :class:`Location` for discussion of the usage of these objects.

       Typically these objects are used to provide more information on how
       a :class:`~ihm.startmodel.StartingModel` was generated, how an
       individual :class:`ihm.protocol.Step` or :class:`ihm.analysis.Step`
       was performed, or to describe the overall modeling (by addition
       to :data:`ihm.System.locations`). This can be useful to capture fine
       details of the modeling that aren't covered by the mmCIF dictionary,
       and to allow models to be precisely reproduced.
    """
    content_type = "Modeling workflow or script"


class VisualizationFileLocation(FileLocation):
    """An externally stored file that is used for visualization.
       See :class:`FileLocation` for a description of the parameters
       and :class:`Location` for discussion of the usage of these objects.
    """
    content_type = "Visualization script"


class Repository(object):
    """A repository containing modeling files, i.e. a collection of related
       files at a remote, public location. This can include code repositories
       such as GitHub, file archival services such as Zenodo, or any other
       service that provides a DOI, such as the supplementary information for
       a publication.

       This can also be used if the script plus related files are part of a
       repository, which has been archived somewhere with a DOI.
       This will be used to construct permanent references to files
       used in this modeling, even if they haven't been uploaded to
       a database such as PDB or EMDB.
       See :meth:`ihm.System.update_locations_in_repositories`.

       See also :class:`FileLocation`.

       :param str doi: the Digital Object Identifier for the repository
       :param str root: the path on the local disk to the top-level
              directory of the repository, or `None` if files in this
              repository aren't checked out.
       :param str url: If given, a location that this repository can be
              downloaded from.
       :param str top_directory: If given, prefix all paths for files in
              this repository with this value. This is useful when the
              archived version of the repository is found in a subdirectory
              at the URL or DOI (for example, GitHub repositories
              archived at Zenodo get placed in a subdirectory named
              for the repository and git hash).
       :param str details: Additional text describing this repository
    """

    reference_type = 'DOI'

    # Two repositories compare equal if their DOIs and URLs are the same
    def __eq__(self, other):
        return self.doi == other.doi and self.url == other.url

    def __hash__(self):
        return hash((self.doi, self.url))

    def __init__(self, doi, root=None, url=None, top_directory=None,
                 details=None):
        # todo: DOI should be optional (could also use URL, local path)
        self.doi = doi
        self.url, self.top_directory = url, top_directory
        self.details = details
        if root is not None:
            # Store absolute path in case the working directory changes later
            self._root = os.path.abspath(root)

    reference = property(lambda self: self.doi)

    def __get_reference_provider(self):
        if 'zenodo' in self.reference:
            return 'Zenodo'
    reference_provider = property(__get_reference_provider)

    def __get_refers_to(self):
        if self.url:
            return 'Archive' if self.url.endswith(".zip") else 'File'
        return 'Other'
    refers_to = property(__get_refers_to)

    @staticmethod
    def _update_in_repos(fileloc, repos):
        """If the given FileLocation maps to somewhere within one of the
           passed repositories, update it to reflect that."""
        if fileloc.repo:
            return
        orig_path = fileloc.path
        for repo in repos:
            relpath = os.path.relpath(orig_path, repo._root)
            if not relpath.startswith('..'):
                # Prefer the shortest paths if multiple repositories can match
                if fileloc.repo is None or len(fileloc.path) > len(relpath):
                    fileloc.repo = repo
                    fileloc.path = relpath

    def _get_full_path(self, path):
        """Prefix the given path with our top-level directory"""
        return os.path.join(self.top_directory or "", path)



"""Classes for handling geometry.

   Geometric objects (see :class:`GeometricObject`) are usually
   used in :class:`~ihm.restraint.GeometricRestraint` objects.
"""


class Center(object):
    """Define the center of a geometric object in Cartesian space.

       :param float x: x coordinate
       :param float y: y coordinate
       :param float z: z coordinate
    """
    def __init__(self, x, y, z):
        self.x, self.y, self.z = x, y, z


class Transformation(object):
    """Rotation and translation applied to an object.

       Transformation objects are typically used in subclasses of
       :class:`GeometricObject`, or by :class:`ihm.dataset.TransformedDataset`.

       :param rot_matrix: Rotation matrix (as a 3x3 array of floats) that
              places the object in its final position.
       :param tr_vector: Translation vector (as a 3-element float list) that
              places the object in its final position.
    """
    def __init__(self, rot_matrix, tr_vector):
        self.rot_matrix, self.tr_vector = rot_matrix, tr_vector

    """Return the identity transformation.

       :return: A new identity Transformation.
       :rtype: :class:`Transformation`
    """
    @classmethod
    def identity(cls):
        return cls([[1., 0., 0.], [0., 1., 0.], [0., 0., 1.]], [0., 0., 0.])


class GeometricObject(object):
    """A generic geometric object. See also :class:`Sphere`, :class:`Torus`,
       :class:`Axis`, :class:`Plane`.

       Geometric objects are typically assigned to one or more
       :class:`~ihm.restraint.GeometricRestraint` objects.

       :param str name: A short user-provided name.
       :param str description: A brief description of the object.
    """
    type = 'other'

    def __init__(self, name=None, description=None):
        self.name, self.description = name, description


class Sphere(GeometricObject):
    """A sphere in Cartesian space.

       :param center: Coordinates of the center of the sphere.
       :type center: :class:`Center`
       :param radius: Radius of the sphere.
       :param transformation: Rotation and translation that moves the sphere
              from the original center to its final location, if any.
       :type transformation: :class:`Transformation`
       :param str name: A short user-provided name.
       :param str description: A brief description of the object.
    """

    type = 'sphere'

    def __init__(self, center, radius, transformation=None,
                 name=None, description=None):
        super(Sphere, self).__init__(name, description)
        self.center, self.transformation = center, transformation
        self.radius = radius


class Torus(GeometricObject):
    """A torus in Cartesian space.

       :param center: Coordinates of the center of the torus.
       :type center: :class:`Center`
       :param major_radius: The major radius - the distance from the center of
              the tube to the center of the torus.
       :param minor_radius: The minor radius - the radius of the tube.
       :param transformation: Rotation and translation that moves the torus
              (which by default lies in the xy plane) from the original
              center to its final location, if any.
       :type transformation: :class:`Transformation`
       :param str name: A short user-provided name.
       :param str description: A brief description of the object.
    """
    type = 'torus'

    def __init__(self, center, major_radius, minor_radius, transformation=None,
                 name=None, description=None):
        super(Torus, self).__init__(name, description)
        self.center, self.transformation = center, transformation
        self.major_radius, self.minor_radius = major_radius, minor_radius


class HalfTorus(GeometricObject):
    """A section of a :class:`Torus`. This is defined as a surface over part
       of the torus with a given thickness, and is often used to represent
       a membrane.

       :param thickness: The thickness of the surface.
       :param inner: True if the surface is the 'inner' half of the torus
              (i.e. closer to the center), False for the outer surface, or
              None for some other section (described in `description`).

       See :class:`Torus` for a description of the other parameters.
    """
    type = 'half-torus'

    def __init__(self, center, major_radius, minor_radius, thickness,
                 transformation=None, inner=None, name=None, description=None):
        super(HalfTorus, self).__init__(name, description)
        self.center, self.transformation = center, transformation
        self.major_radius, self.minor_radius = major_radius, minor_radius
        self.thickness, self.inner = thickness, inner


class Axis(GeometricObject):
    """One of the three Cartesian axes - see :class:`XAxis`, :class:`YAxis`,
       :class:`ZAxis`.

       :param transformation: Rotation and translation that moves the axis
              from the original Cartesian axis to its final location, if any.
       :type transformation: :class:`Transformation`
       :param str name: A short user-provided name.
       :param str description: A brief description of the object.
    """
    type = 'axis'

    def __init__(self, transformation=None, name=None, description=None):
        super(Axis, self).__init__(name, description)
        self.transformation = transformation


class XAxis(Axis):
    """The x Cartesian axis.

       See :class:`GeometricObject` for a description of the parameters.
    """
    axis_type = 'x-axis'


class YAxis(Axis):
    """The y Cartesian axis.

       See :class:`GeometricObject` for a description of the parameters.
    """
    axis_type = 'y-axis'


class ZAxis(Axis):
    """The z Cartesian axis.

       See :class:`GeometricObject` for a description of the parameters.
    """
    axis_type = 'z-axis'


class Plane(GeometricObject):
    """A plane in Cartesian space - see :class:`XYPlane`, :class:`YZPlane`,
       :class:`XZPlane`.

       :param transformation: Rotation and translation that moves the plane
              from the original position to its final location, if any.
       :type transformation: :class:`Transformation`
       :param str name: A short user-provided name.
       :param str description: A brief description of the object.
    """
    type = 'plane'

    def __init__(self, transformation=None, name=None, description=None):
        super(Plane, self).__init__(name, description)
        self.transformation = transformation


class XYPlane(Plane):
    """The xy plane in Cartesian space.

       See :class:`GeometricObject` for a description of the parameters.
    """
    plane_type = 'xy-plane'


class YZPlane(Plane):
    """The yz plane in Cartesian space.

       See :class:`GeometricObject` for a description of the parameters.
    """
    plane_type = 'yz-plane'


class XZPlane(Plane):
    """The xz plane in Cartesian space.

       See :class:`GeometricObject` for a description of the parameters.
    """
    plane_type = 'xz-plane'



"""Classes for providing extra information about an :class:`ihm.Entity`"""

# Handle different naming of urllib in Python 2/3
try:
    import urllib.request as urllib2
except ImportError:
    import urllib2
import sys


class Reference(object):
    """Base class for extra information about an :class:`ihm.Entity`.

       This class is not used directly; instead, use a subclass such as
       :class:`Sequence` or :class:`UniProtSequence`. These objects are
       then typically passed to the :class:`ihm.Entity` constructor."""
    pass


class Sequence(Reference):
    """Point to the sequence of an :class:`ihm.Entity` in a sequence database;
       convenience subclasses are provided for common sequence databases such
       as :class:`UniProtSequence`.

       These objects are typically passed to the :class:`ihm.Entity`
       constructor.

       See also :attr:`alignments` to describe the correspondence between
       the database and entity sequences.

       :param str db_name: The name of the database.
       :param str db_code: The name of the sequence in the database.
       :param str accession: The database accession.
       :param str sequence: The complete sequence, as a string of
              one-letter codes.
       :param str details: Longer text describing the sequence.
    """

    def __init__(self, db_name, db_code, accession, sequence, details=None):
        self.db_name, self.db_code = db_name, db_code
        self.accession = accession
        self.sequence, self.details = sequence, details

        #: All alignments between the reference and entity sequences, as
        #: :class:`Alignment` objects. If none are provided, a simple 1:1
        #: alignment is assumed.
        self.alignments = []

    def _get_alignments(self):
        if self.alignments:
            return self.alignments
        elif not hasattr(self, '_default_alignment'):
            self._default_alignment = Alignment()
        return [self._default_alignment]


class UniProtSequence(Sequence):
    """Point to the sequence of an :class:`ihm.Entity` in UniProt.

       These objects are typically passed to the :class:`ihm.Entity`
       constructor.

       :param str db_code: The UniProt name (e.g. NUP84_YEAST)
       :param str accession: The UniProt accession (e.g. P52891)

       See :class:`Sequence` for a description of the remaining parameters.
    """

    _db_name = 'UNP'

    def __init__(self, db_code, accession, sequence, details=None):
        super(UniProtSequence, self).__init__(
            self._db_name, db_code, accession, sequence, details)

    def __str__(self):
        return "<ihm.reference.UniProtSequence(%s)>" % self.accession

    @classmethod
    def from_accession(cls, accession):
        """Create :class:`UniProtSequence` from just an accession.
           This is done by querying the UniProt web API, so requires network
           access.

           :param str accession: The UniProt accession (e.g. P52891)
        """
        # urlopen returns bytes
        if sys.version_info[0] >= 3:
            def decode(t):
                return t.decode('ascii')
        else:
            def decode(t):
                return t
        url = 'https://www.uniprot.org/uniprot/%s.fasta' % accession
        with urllib2.urlopen(url) as fh:
            header = decode(fh.readline())
            spl = header.split('|')
            if len(spl) < 3 or spl[0] not in ('>sp', '>tr'):
                raise ValueError("Cannot parse UniProt header %s" % header)
            cd = spl[2].split(None, 1)
            code = cd[0]
            details = cd[1].rstrip('\r\n') if len(cd) > 1 else None
            seq = decode(fh.read()).replace('\n', '')
            return cls(code, accession, seq, details)


class Alignment(object):
    """A sequence range that aligns between the database and the entity.
       This describes part of the sequence in the sequence database
       (:class:`Sequence`) and in the :class:`ihm.Entity`. The two ranges
       must be the same length and have the same primary sequence (any
       differences must be described with :class:`SeqDif` objects).

       :param int db_begin: The first residue in the database sequence
              that is used (defaults to the entire sequence).
       :param int db_end: The last residue in the database sequence
              that is used (or None, the default, to use the entire sequence).
       :param int entity_begin: The first residue in the :class:`~ihm.Entity`
              sequence that is taken from the reference (defaults to the entire
              entity sequence).
       :param int entity_end: The last residue in the :class:`~ihm.Entity`
              sequence that is taken from the reference (or None, the default,
              to use the entire sequence).
       :param seq_dif: Single-point mutations made to the sequence.
       :type seq_dif: Sequence of :class:`SeqDif` objects.
    """
    def __init__(self, db_begin=1, db_end=None, entity_begin=1,
                 entity_end=None, seq_dif=[]):
        self.db_begin, self.db_end = db_begin, db_end
        self.entity_begin, self.entity_end = entity_begin, entity_end
        self.seq_dif = []
        self.seq_dif.extend(seq_dif)


class SeqDif(object):
    """Annotate a sequence difference between a reference and entity sequence.
       See :class:`Alignment`.

       :param int seq_id: The residue index in the entity sequence.
       :param db_monomer: The monomer type (as a :class:`~ihm.ChemComp` object)
              in the reference sequence.
       :type db_monomer: :class:`ihm.ChemComp`
       :param monomer: The monomer type (as a :class:`~ihm.ChemComp` object)
              in the entity sequence.
       :type monomer: :class:`ihm.ChemComp`
       :param str details: Descriptive text for the sequence difference.
    """
    def __init__(self, seq_id, db_monomer, monomer, details=None):
        self.seq_id, self.db_monomer = seq_id, db_monomer
        self.monomer, self.details = monomer, details



# coding=utf-8

"""Classes for handling experimental datasets used by mmCIF models.
"""


class Dataset(object):
    """A set of input data, for example, a crystal structure or EM map.

       :param location: a pointer to where the
              dataset is stored. This is usually a subclass of
              :class:`~ihm.location.DatabaseLocation` if the dataset is
              deposited in a database such as PDB or EMDB, or
              :class:`~ihm.location.InputFileLocation` if the dataset is stored
              in an external file.
       :type location: :class:`ihm.location.Location`
       :param str details: Text giving more information about the dataset.
    """

    _eq_keys = ['location']

    # Datasets compare equal iff they are the same class and have the
    # same attributes
    def _eq_vals(self):
        return tuple([self.__class__]
                     + [getattr(self, x) for x in self._eq_keys])

    def __eq__(self, other):
        return self._eq_vals() == other._eq_vals()

    def __hash__(self):
        return hash(self._eq_vals())

    data_type = 'Other'

    def __init__(self, location, details=None):
        self.location, self.details = location, details

        #: A list of :class:`Dataset` and/or :class:`TransformedDataset`
        #: objects from which this one was derived.
        #: For example, a 3D EM map may be derived from a set of 2D images.
        self.parents = []

    def add_primary(self, dataset):
        """Add another Dataset from which this one was ultimately derived,
           i.e. it is added as a parent, unless a parent already exists,
           in which case it is added as a grandparent, and so on."""
        root = self
        while root.parents:
            if len(root.parents) > 1:
                raise ValueError("This dataset has multiple parents - don't "
                                 "know which one to add to")
            root = root.parents[0]
        root.parents.append(dataset)


class TransformedDataset(object):
    """A :class:`Dataset` that should be rotated or translated before using.
       This is typically used for derived datasets
       (see :attr:`Dataset.parents`) where the derived dataset lies in a
       different dataset from the parent (for example, it was moved to better
       align with the model's reference frame or other experimental data).
       The transformation that places the derived dataset on the parent
       is recorded here.

       :param dataset: The (parent) dataset.
       :type dataset: :class:`Dataset`
       :param transform: The rotation and translation that places a
              derived dataset on this dataset.
       :type transform: :class:`ihm.geometry.Transformation`
    """
    def __init__(self, dataset, transform):
        self.dataset, self.transform = dataset, transform


class DatasetGroup(list):
    """A set of :class:`Dataset` objects that are handled together.
       This is implemented as a simple list.

       :param sequence elements: Initial set of datasets.
       :param str name: Short text name of this group.
       :param str application: Text that shows how this group is used.
       :param str details: Longer text that describes this group.

       Normally a group is passed to one or more
       :class:`~ihm.protocol.Protocol` or :class:`~ihm.analysis.Analysis`
       objects, although unused groups can still be included in the file
       if desired by adding them to :attr:`ihm.System.orphan_dataset_groups`.
    """

    # For backwards compatibility with earlier versions of this class which
    # didn't specify name/application/details
    name = application = details = None

    def __init__(self, elements=(), name=None, application=None, details=None):
        super(DatasetGroup, self).__init__(elements)
        self.name, self.application = name, application
        self.details = details


class CXMSDataset(Dataset):
    """Processed cross-links from a CX-MS experiment"""
    data_type = 'CX-MS data'


class MassSpecDataset(Dataset):
    """Raw mass spectrometry files such as peaklists"""
    data_type = 'Mass Spectrometry data'


class HDXDataset(Dataset):
    """Data from a hydrogen/deuterium exchange experiment"""
    data_type = 'H/D exchange data'


class PDBDataset(Dataset):
    """An experimentally-determined 3D structure as a set of a coordinates,
       usually in a PDB file"""
    data_type = 'Experimental model'


class ComparativeModelDataset(Dataset):
    """A 3D structure determined by comparative modeling"""
    data_type = 'Comparative model'


class IntegrativeModelDataset(Dataset):
    """A 3D structure determined by integrative modeling"""
    data_type = 'Integrative model'


class DeNovoModelDataset(Dataset):
    """A 3D structure determined by de novo modeling"""
    data_type = 'De Novo model'


class NMRDataset(Dataset):
    """A nuclear magnetic resonance (NMR) dataset"""
    data_type = 'NMR data'


class MutagenesisDataset(Dataset):
    """Mutagenesis data"""
    data_type = 'Mutagenesis data'


class EMDensityDataset(Dataset):
    """A 3D electron microscopy dataset"""
    data_type = '3DEM volume'


class EMMicrographsDataset(Dataset):
    """Raw 2D electron micrographs"""
    data_type = 'EM raw micrographs'


class EM2DClassDataset(Dataset):
    """2DEM class average"""
    data_type = '2DEM class average'


class SASDataset(Dataset):
    """SAS data"""
    data_type = 'SAS data'


class FRETDataset(Dataset):
    """Data from a Förster resonance energy transfer (FRET) experiment"""
    data_type = 'Single molecule FRET data'


class YeastTwoHybridDataset(Dataset):
    """Yeast two-hybrid data"""
    data_type = 'Yeast two-hybrid screening data'


class GeneticInteractionsDataset(Dataset):
    """Quantitative measurements of genetic interactions"""
    data_type = 'Quantitative measurements of genetic interactions'



# -*- coding: utf-8 -*-

"""Citations for some commonly-used software packages.

   Each of these is an instance of the :class:`ihm.Citation` class,
   and so can be used anywhere these objects are required, generally for
   :class:`ihm.Software`.
"""

import ihm


imp = ihm.Citation(
    pmid='22272186',
    title='Putting the pieces together: integrative modeling platform '
          'software for structure determination of macromolecular assemblies',
    journal='PLoS Biol', volume=10, page_range='e1001244', year=2012,
    authors=['Russel D', 'Lasker K', 'Webb B', 'Velázquez-Muriel J', 'Tjioe E',
             'Schneidman-Duhovny D', 'Peterson B', 'Sali A'],
    doi='10.1371/journal.pbio.1001244')


pmi = ihm.Citation(
    pmid='31396911',
    title='Modeling Biological Complexes Using Integrative Modeling Platform.',
    journal='Methods Mol Biol', volume=2022, page_range=(353, 377), year=2019,
    authors=['Saltzberg D', 'Greenberg CH', 'Viswanath S', 'Chemmama I',
             'Webb B', 'Pellarin R', 'Echeverria I', 'Sali A'],
    doi='10.1007/978-1-4939-9608-7_15')


modeller = ihm.Citation(
    pmid='8254673',
    title='Comparative protein modelling by satisfaction of '
          'spatial restraints.',
    journal='J Mol Biol', volume=234, page_range=(779, 815), year=1993,
    authors=['Sali A', 'Blundell TL'], doi='10.1006/jmbi.1993.1626')


psipred = ihm.Citation(
    pmid='10493868',
    title='Protein secondary structure prediction based on position-specific '
          'scoring matrices.',
    journal='J Mol Biol', volume=292, page_range=(195, 202), year=1999,
    authors=['Jones DT'], doi='10.1006/jmbi.1999.3091')


disopred = ihm.Citation(
    pmid='25391399',
    title='DISOPRED3: precise disordered region predictions with annotated '
          'protein-binding activity.',
    journal='Bioinformatics', volume=31, page_range=(857, 863), year=2015,
    authors=['Jones DT', 'Cozzetto D'], doi='10.1093/bioinformatics/btu744')


hhpred = ihm.Citation(
    pmid='15980461',
    title='The HHpred interactive server for protein homology detection '
          'and structure prediction.',
    journal='Nucleic Acids Res', volume=33, page_range=('W244', 'W248'),
    year=2005, authors=['Söding J', 'Biegert A', 'Lupas AN'],
    doi='10.1093/nar/gki408')


relion = ihm.Citation(
    pmid='23000701',
    title='RELION: implementation of a Bayesian approach to cryo-EM '
          'structure determination.',
    journal='J Struct Biol', volume=180, page_range=(519, 530), year=2012,
    authors=['Scheres SH'], doi='10.1016/j.jsb.2012.09.006')


phyre2 = ihm.Citation(
    pmid='25950237',
    title='The Phyre2 web portal for protein modeling, prediction '
          'and analysis.',
    journal='Nat Protoc', volume=10, page_range=('845', '858'), year=2015,
    authors=['Kelley LA', 'Mezulis S', 'Yates CM', 'Wass MN', 'Sternberg MJ'],
    doi='10.1038/nprot.2015.053')


swiss_model = ihm.Citation(
    pmid='29788355',
    title='SWISS-MODEL: homology modelling of protein structures '
          'and complexes.',
    journal='Nucleic Acids Res', volume=46, page_range=('W296', 'W303'),
    year=2018,
    authors=['Waterhouse A', 'Bertoni M', 'Bienert S', 'Studer G',
             'Tauriello G', 'Gumienny R', 'Heer FT', 'de Beer TAP',
             'Rempfer C', 'Bordoli L', 'Lepore R', 'Schwede T'],
    doi='10.1093/nar/gky427')



"""Classes for handling restraints on the system.
"""

import ihm


class PseudoSite(object):
    """Selection of a pseudo position in the system.
       Pseudo positions are typically used to reference a point or sphere
       that is not explcitly represented, in a :class:`PseudoSiteFeature`
       or :class:`CrossLinkPseudoSite`.

       :param float x: Cartesian X coordinate of this site.
       :param float y: Cartesian Y coordinate of this site.
       :param float z: Cartesian Z coordinate of this site.
       :param float radius: Radius of the site, if applicable.
       :param str description: Additional text describing this feature.
    """

    def __init__(self, x, y, z, radius=None, description=None):
        self.x, self.y, self.z = x, y, z
        self.radius = radius
        self.description = description


class Restraint(object):
    """Base class for all restraints.
       See :attr:`ihm.System.restraints`.
    """
    pass


class RestraintGroup(list):
    """A set of related :class:`Restraint` objects.
       This is implemented as a simple list.

       Note that due to limitations of the underlying dictionary, only
       certain combinations of restraints can be placed in groups.
       In particular, all objects in a group must be of the same type, and
       only certain types (currently only :class:`DerivedDistanceRestraint`
       and :class:`PredictedContactRestraint`) can be grouped.

       Empty groups can be created, but will be ignored on output as the
       dictionary does not support them.

       Restraint groups should be stored in the system by adding them to
       :attr:`ihm.System.restraint_groups`.
    """
    pass


class EM3DRestraint(Restraint):
    """Restrain part of the system to match an electron microscopy density map.

       :param dataset: Reference to the density map data (usually
              an :class:`~ihm.dataset.EMDensityDataset`).
       :type dataset: :class:`~ihm.dataset.Dataset`
       :param assembly: The part of the system that is fit into the map.
       :type assembly: :class:`~ihm.Assembly`
       :param bool segment: True iff the map has been segmented.
       :param str fitting_method: The method used to fit the model
              into the map.
       :param fitting_method_citation: The publication describing the fitting
              method.
       :type fitting_method_citation: :class:`~ihm.Citation`
       :param int number_of_gaussians: Number of Gaussians used to represent
              the map as a Gaussian Mixture Model (GMM), if applicable.
       :param str details: Additional details regarding the fitting.
    """

    def __init__(self, dataset, assembly, segment=None, fitting_method=None,
                 fitting_method_citation=None, number_of_gaussians=None,
                 details=None):
        self.dataset, self.assembly = dataset, assembly
        self.segment, self.fitting_method = segment, fitting_method
        self.fitting_method_citation = fitting_method_citation
        self.number_of_gaussians = number_of_gaussians
        self.details = details

        #: Information about the fit of each model to this restraint's data.
        #: This is a Python dict where keys are :class:`~ihm.model.Model`
        #: objects and values are :class:`EM3DRestraintFit` objects.
        self.fits = {}


class EM3DRestraintFit(object):
    """Information on the fit of a model to an :class:`EM3DRestraint`.
       See :attr:`EM3DRestaint.fits`.

       :param float cross_correlation_coefficient: The fit between the model
              and the map.
    """
    __slots__ = ["cross_correlation_coefficient"]  # Reduce memory usage

    def __init__(self, cross_correlation_coefficient=None):
        self.cross_correlation_coefficient = cross_correlation_coefficient


class SASRestraint(Restraint):
    """Restrain part of the system to match small angle scattering (SAS) data.

       :param dataset: Reference to the SAS data (usually
              an :class:`~ihm.dataset.SASDataset`).
       :type dataset: :class:`~ihm.dataset.Dataset`
       :param assembly: The part of the system that is fit against SAS data.
       :type assembly: :class:`~ihm.Assembly`
       :param bool segment: True iff the SAS profile has been segmented.
       :param str fitting_method: The method used to fit the model against the
              SAS data (e.g. FoXS, DAMMIF).
       :param str fitting_atom_type: The set of atoms fit against the data
              (e.g. "Heavy atoms", "All atoms").
       :param bool multi_state: Whether multiple state fitting was done.
       :param float radius_of_gyration: Radius of gyration obtained from the
              SAS profile, if used as part of the restraint.
       :param str details: Additional details regarding the fitting.
    """

    def __init__(self, dataset, assembly, segment=None, fitting_method=None,
                 fitting_atom_type=None, multi_state=None,
                 radius_of_gyration=None, details=None):
        self.dataset, self.assembly = dataset, assembly
        self.segment, self.fitting_method = segment, fitting_method
        self.fitting_atom_type = fitting_atom_type
        self.multi_state = multi_state
        self.radius_of_gyration = radius_of_gyration
        self.details = details

        #: Information about the fit of each model to this restraint's data.
        #: This is a Python dict where keys are :class:`~ihm.model.Model`
        #: objects and values are :class:`SASRestraintFit` objects.
        self.fits = {}


class SASRestraintFit(object):
    """Information on the fit of a model to a :class:`SASRestraint`.
       See :attr:`SASRestaint.fits`.

       :param float chi_value: The fit between the model and the SAS data.
    """
    __slots__ = ["chi_value"]  # Reduce memory usage

    def __init__(self, chi_value=None):
        self.chi_value = chi_value


class EM2DRestraint(Restraint):
    """Restrain part of the system to match an electron microscopy class
       average.

       :param dataset: Reference to the class average data (usually
              an :class:`~ihm.dataset.EM2DClassDataset`).
       :type dataset: :class:`~ihm.dataset.Dataset`
       :param assembly: The part of the system that is fit against the class.
       :type assembly: :class:`~ihm.Assembly`
       :param bool segment: True iff the image has been segmented.
       :param int number_raw_micrographs: The number of particles picked from
              the original raw micrographs that were used to create the
              class average.
       :param float pixel_size_width: Width of each pixel in the image, in
              angstroms.
       :param float pixel_size_height: Height of each pixel in the image, in
              angstroms.
       :param float image_resolution: Resolution of the image, in angstroms.
       :param int number_of_projections: Number of projections of the assembly
              used to fit against the image, if applicable.
       :param str details: Additional details regarding the fitting.
    """

    def __init__(self, dataset, assembly, segment=None,
                 number_raw_micrographs=None, pixel_size_width=None,
                 pixel_size_height=None, image_resolution=None,
                 number_of_projections=None, details=None):
        self.dataset, self.assembly = dataset, assembly
        self.segment = segment
        self.number_raw_micrographs = number_raw_micrographs
        self.pixel_size_width = pixel_size_width
        self.pixel_size_height = pixel_size_height
        self.image_resolution = image_resolution
        self.number_of_projections = number_of_projections
        self.details = details

        #: Information about the fit of each model to this restraint's data.
        #: This is a Python dict where keys are :class:`~ihm.model.Model`
        #: objects and values are :class:`EM2DRestraintFit` objects.
        self.fits = {}


class EM2DRestraintFit(object):
    """Information on the fit of a model to an :class:`EM2DRestraint`.
       See :attr:`EM2DRestaint.fits`.

       :param float cross_correlation_coefficient: The fit between the model
              and the class average.
       :param rot_matrix: Rotation matrix (as a 3x3 array of floats) that
              places the model on the image.
       :param tr_vector: Translation vector (as a 3-element float list) that
              places the model on the image.
    """
    __slots__ = ["cross_correlation_coefficient",
                 "rot_matrix", "tr_vector"]  # Reduce memory usage

    def __init__(self, cross_correlation_coefficient=None,
                 rot_matrix=None, tr_vector=None):
        self.cross_correlation_coefficient = cross_correlation_coefficient
        self.rot_matrix, self.tr_vector = rot_matrix, tr_vector


class CrossLinkRestraint(Restraint):
    """Restrain part of the system to match a set of cross-links.

       :param dataset: Reference to the cross-link data (usually
              a :class:`~ihm.dataset.CXMSDataset`).
       :type dataset: :class:`~ihm.dataset.Dataset`
       :param linker: The type of chemical linker used.
       :type linker: :class:`ihm.ChemDescriptor`
    """

    assembly = None  # no struct_assembly_id for XL restraints

    def __init__(self, dataset, linker):
        self.dataset, self.linker = dataset, linker

        #: All cross-links identified in the experiment, as a simple list
        #: of lists of :class:`ExperimentalCrossLink` objects. All cross-links
        #: in the same sublist are treated as experimentally ambiguous. For
        #: example, xl2 and xl3 here are considered ambiguous::
        #:
        #:     restraint.experimental_cross_links.append([xl1])
        #:     restraint.experimental_cross_links.append([xl2, xl3])
        self.experimental_cross_links = []

        #: All cross-links used in the modeling, as a list of
        #: :class:`CrossLink` objects.
        self.cross_links = []


class ExperimentalCrossLink(object):
    """A cross-link identified in the experiment.

       :param residue1: The first residue linked by the cross-link.
       :type residue1: :class:`ihm.Residue`
       :param residue2: The second residue linked by the cross-link.
       :type residue2: :class:`ihm.Residue`
       :param str details: Additional text describing the cross-link.
    """
    def __init__(self, residue1, residue2, details=None):
        self.residue1, self.residue2 = residue1, residue2
        self.details = details


class DistanceRestraint(object):
    """Abstract base class for all distance restraints. These are typically
       used in a :class:`DerivedDistanceRestraint`.

       Do not use this class directly but instead use a derived class
       such as :class:`HarmonicDistanceRestraint`,
       :class:`UpperBoundDistanceRestraint`,
       :class:`LowerBoundDistanceRestraint`,
       or :class:`LowerUpperBoundDistanceRestraint`.
    """

    #: The minimum distance allowed for this restraint,
    #: or None if unconstrained
    distance_lower_limit = None

    #: The maximum distance allowed for this restraint,
    #: or None if unconstrained
    distance_upper_limit = None


class HarmonicDistanceRestraint(DistanceRestraint):
    """Harmonically restrain two objects to be close to a given distance apart.
       These objects are typically used in a :class:`DerivedDistanceRestraint`.

       :param float distance: Equilibrium distance
    """
    restraint_type = 'harmonic'

    def __init__(self, distance):
        self.distance = distance

    #: The equilibrium distance
    distance_lower_limit = property(lambda self: self.distance)

    #: The equilibrium distance
    distance_upper_limit = distance_lower_limit


class UpperBoundDistanceRestraint(DistanceRestraint):
    """Harmonically restrain two objects to be below a given distance apart.
       These objects are typically used in a :class:`DerivedDistanceRestraint`.

       :param float distance: Distance threshold
    """
    restraint_type = 'upper bound'

    def __init__(self, distance):
        self.distance = distance

    #: The maximum distance allowed by this restraint
    distance_upper_limit = property(lambda self: self.distance)

    #: Minimum distance (unconstrained, so always None)
    distance_lower_limit = None


class LowerBoundDistanceRestraint(DistanceRestraint):
    """Harmonically restrain two objects to be above a given distance apart.
       These objects are typically used in a :class:`DerivedDistanceRestraint`.

       :param float distance: Distance threshold
    """
    restraint_type = 'lower bound'

    def __init__(self, distance):
        self.distance = distance

    #: The minimum distance allowed by this restraint
    distance_lower_limit = property(lambda self: self.distance)

    #: Maximum distance (unconstrained, so always None)
    distance_upper_limit = None


class LowerUpperBoundDistanceRestraint(DistanceRestraint):
    """Harmonically restrain two objects to be above a given distance
       and below another distance apart.
       These objects are typically used in a :class:`DerivedDistanceRestraint`.

       :param float distance_lower_limit: Lower bound on the distance.
       :param float distance_upper_limit: Upper bound on the distance.
    """
    restraint_type = 'lower and upper bound'

    def __init__(self, distance_lower_limit, distance_upper_limit):
        #: The minimum distance allowed by this restraint
        self.distance_lower_limit = distance_lower_limit

        #: The maximum distance allowed by this restraint
        self.distance_upper_limit = distance_upper_limit


class CrossLink(object):
    """Base class for all cross-links used in the modeling.
       Do not use this class directly, but instead use a subclass:
       :class:`ResidueCrossLink`, :class:`AtomCrossLink`,
       or :class:`FeatureCrossLink`."""
    pass


class CrossLinkPseudoSite(object):
    """Pseudo site corresponding to one end of a cross-link.

       These objects are used when the end of a cross-link is not represented
       in the model but its position is known (e.g. it may have been
       approximated given the position of nearby residues). They are passed
       as the ``pseudo1`` or ``pseudo2`` arguments to :class:`CrossLink`
       subclasses.

       :param site: The pseudo site coordinates
       :type site: :class:`PseudoSite`
       :param model: The model in whose coordinate system the pseudo site
              is active (if not specified, the coordinates are assumed to
              be valid for all models using this cross-link).
       :type model: :class:`ihm.model.Model`
    """
    def __init__(self, site, model=None):
        self.site, self.model = site, model


class ResidueCrossLink(CrossLink):
    """A cross-link used in the modeling, applied to residue alpha carbon atoms.

       :param experimental_cross_link: The corresponding cross-link identified
              by experiment. Multiple cross-links can map to a single
              experimental identification.
       :type experimental_cross_link: :class:`ExperimentalCrossLink`
       :param asym1: The asymmetric unit containing the first linked residue.
       :type asym1: :class:`ihm.AsymUnit`
       :param asym2: The asymmetric unit containing the second linked residue.
       :type asym2: :class:`ihm.AsymUnit`
       :param distance: Restraint on the distance.
       :type distance: :class:`DistanceRestraint`
       :param float psi: Initial uncertainty in the experimental data.
       :param float sigma1: Initial uncertainty in the position of the first
              residue.
       :param float sigma2: Initial uncertainty in the position of the second
              residue.
       :param bool restrain_all: If True, all cross-links are restrained.
       :param pseudo1: List of pseudo sites representing the position of the
              first residue (if applicable).
       :type pseudo1: List of :class:`CrossLinkPseudoSite`
       :param pseudo2: List of pseudo sites representing the position of the
              second residue (if applicable).
       :type pseudo2: List of :class:`CrossLinkPseudoSite`
    """
    granularity = 'by-residue'
    atom1 = atom2 = None

    def __init__(self, experimental_cross_link, asym1, asym2, distance,
                 psi=None, sigma1=None, sigma2=None, restrain_all=None,
                 pseudo1=None, pseudo2=None):
        self.experimental_cross_link = experimental_cross_link
        self.asym1, self.asym2 = asym1, asym2
        self.psi, self.sigma1, self.sigma2 = psi, sigma1, sigma2
        self.distance, self.restrain_all = distance, restrain_all
        self.pseudo1, self.pseudo2 = pseudo1, pseudo2

        #: Information about the fit of each model to this cross-link
        #: This is a Python dict where keys are :class:`~ihm.model.Model`
        #: objects and values are :class:`CrossLinkFit` objects.
        self.fits = {}


class FeatureCrossLink(CrossLink):
    """A cross-link used in the modeling, applied to the closest primitive
       object with the highest resolution.

       :param experimental_cross_link: The corresponding cross-link identified
              by experiment. Multiple cross-links can map to a single
              experimental identification.
       :type experimental_cross_link: :class:`ExperimentalCrossLink`
       :param asym1: The asymmetric unit containing the first linked residue.
       :type asym1: :class:`ihm.AsymUnit`
       :param asym2: The asymmetric unit containing the second linked residue.
       :type asym2: :class:`ihm.AsymUnit`
       :param distance: Restraint on the distance.
       :type distance: :class:`DistanceRestraint`
       :param float psi: Initial uncertainty in the experimental data.
       :param float sigma1: Initial uncertainty in the position of the first
              residue.
       :param float sigma2: Initial uncertainty in the position of the second
              residue.
       :param bool restrain_all: If True, all cross-links are restrained.
       :param pseudo1: List of pseudo sites representing the position of the
              first residue (if applicable).
       :type pseudo1: List of :class:`CrossLinkPseudoSite`
       :param pseudo2: List of pseudo sites representing the position of the
              second residue (if applicable).
       :type pseudo2: List of :class:`CrossLinkPseudoSite`
    """
    granularity = 'by-feature'
    atom1 = atom2 = None

    def __init__(self, experimental_cross_link, asym1, asym2, distance,
                 psi=None, sigma1=None, sigma2=None, restrain_all=None,
                 pseudo1=None, pseudo2=None):
        self.experimental_cross_link = experimental_cross_link
        self.asym1, self.asym2 = asym1, asym2
        self.psi, self.sigma1, self.sigma2 = psi, sigma1, sigma2
        self.distance, self.restrain_all = distance, restrain_all
        self.pseudo1, self.pseudo2 = pseudo1, pseudo2

        #: Information about the fit of each model to this cross-link
        #: This is a Python dict where keys are :class:`~ihm.model.Model`
        #: objects and values are :class:`CrossLinkFit` objects.
        self.fits = {}


class AtomCrossLink(CrossLink):
    """A cross-link used in the modeling, applied to the specified atoms.

       :param experimental_cross_link: The corresponding cross-link identified
              by experiment. Multiple cross-links can map to a single
              experimental identification.
       :type experimental_cross_link: :class:`ExperimentalCrossLink`
       :param asym1: The asymmetric unit containing the first linked residue.
       :type asym1: :class:`ihm.AsymUnit`
       :param asym2: The asymmetric unit containing the second linked residue.
       :type asym2: :class:`ihm.AsymUnit`
       :param str atom1: The name of the first linked atom.
       :param str atom2: The name of the second linked atom.
       :param distance: Restraint on the distance.
       :type distance: :class:`DistanceRestraint`
       :param float psi: Initial uncertainty in the experimental data.
       :param float sigma1: Initial uncertainty in the position of the first
              residue.
       :param float sigma2: Initial uncertainty in the position of the second
              residue.
       :param bool restrain_all: If True, all cross-links are restrained.
       :param pseudo1: List of pseudo sites representing the position of the
              first residue (if applicable).
       :type pseudo1: List of :class:`CrossLinkPseudoSite`
       :param pseudo2: List of pseudo sites representing the position of the
              second residue (if applicable).
       :type pseudo2: List of :class:`CrossLinkPseudoSite`
    """
    granularity = 'by-atom'

    def __init__(self, experimental_cross_link, asym1, asym2, atom1, atom2,
                 distance, psi=None, sigma1=None, sigma2=None,
                 restrain_all=None, pseudo1=None, pseudo2=None):
        self.experimental_cross_link = experimental_cross_link
        self.asym1, self.asym2 = asym1, asym2
        self.atom1, self.atom2 = atom1, atom2
        self.psi, self.sigma1, self.sigma2 = psi, sigma1, sigma2
        self.distance, self.restrain_all = distance, restrain_all
        self.pseudo1, self.pseudo2 = pseudo1, pseudo2

        #: Information about the fit of each model to this cross-link
        #: This is a Python dict where keys are :class:`~ihm.model.Model`
        #: objects and values are :class:`CrossLinkFit` objects.
        self.fits = {}


class CrossLinkFit(object):
    """Information on the fit of a model to a :class:`CrossLink`.
       See :attr:`ResidueCrossLink.fits`, :attr:`AtomCrossLink.fits`, or
       :attr:`FeatureCrossLink.fits`.

       :param float psi: Uncertainty in the experimental data.
       :param float sigma1: Uncertainty in the position of the first residue.
       :param float sigma2: Uncertainty in the position of the second residue.
    """
    __slots__ = ["psi", "sigma1", "sigma2"]  # Reduce memory usage

    def __init__(self, psi=None, sigma1=None, sigma2=None):
        self.psi, self.sigma1, self.sigma2 = psi, sigma1, sigma2


class Feature(object):
    """Base class for selecting parts of the system that a restraint acts on.
       See :class:`ResidueFeature`, :class:`AtomFeature`,
       :class:`NonPolyFeature`, and :class:`PseudoSiteFeature`.

       Features are typically assigned to one or more
       :class:`~ihm.restraint.GeometricRestraint` or
       :class:`~ihm.restraint.DerivedDistanceRestraint` objects.
    """
    details = None

    def _all_entities_or_asyms(self):
        # Get all Entities or AsymUnits referenced by this object
        return []


class ResidueFeature(Feature):
    """Selection of one or more residues from the system.

       Residues can be selected from both :class:`AsymUnit` and
       :class:`Entity` (the latter implies that it selects residues
       in all instances of that entity). Individual residues can
       also be selected by passing :class:`Residue` objects.

       :param sequence ranges: A list of :class:`AsymUnitRange`,
              :class:`AsymUnit`, :class:`EntityRange`, :class:`Residue`,
              and/or :class:`Entity` objects.
       :param str details: Additional text describing this feature.
    """

    # Type is 'residue' if each range selects a single residue, otherwise
    # it is 'residue range'
    def __get_type(self):
        for r in self.ranges:
            if r.seq_id_range[0] != r.seq_id_range[1]:
                return 'residue range'
        return 'residue'
    type = property(__get_type)

    def __init__(self, ranges, details=None):
        self.ranges, self.details = ranges, details
        _ = self._get_entity_type()

    def _all_entities_or_asyms(self):
        return self.ranges

    def _get_entity_type(self):
        def _get_entity(x):
            if isinstance(x, ihm.Entity):
                return x
            return x.entity if x.entity else x.asym.entity
        if any(not _get_entity(r).is_polymeric() for r in self.ranges):
            raise ValueError("%s cannot select non-polymeric entities" % self)
        else:
            return _get_entity(self.ranges[0]).type if self.ranges else None


class AtomFeature(Feature):
    """Selection of one or more atoms from the system.
       Atoms can be selected from polymers or non-polymers (but not both).
       Atoms can also be selected from both :class:`AsymUnit` and
       :class:`Entity` (the latter implies that it selects atoms
       in all instances of that entity).
       For selecting an entire polymer or residue(s),
       see :class:`ResidueFeature`. For selecting an entire non-polymer,
       see :class:`NonPolyFeature`.

       :param sequence atoms: A list of :class:`ihm.Atom` objects.
       :param str details: Additional text describing this feature.
    """
    type = 'atom'

    def __init__(self, atoms, details=None):
        self.atoms, self.details = atoms, details
        _ = self._get_entity_type()

    def _get_entity_type(self):
        def _get_entity(residue):
            return residue.entity if residue.entity else residue.asym.entity
        types = frozenset(_get_entity(a.residue).type for a in self.atoms)
        if len(types) > 1:
            raise ValueError("%s cannot span both polymeric and "
                             "non-polymeric entities" % self)
        elif types:
            return tuple(types)[0]


class NonPolyFeature(Feature):
    """Selection of one or more non-polymers from the system.
       To select individual atoms from a non-polymer, see :class:`AtomFeature`.

       Features can include both :class:`AsymUnit` and
       :class:`Entity` (the latter implies that it selects non-polymers
       in all instances of that entity).

       :param sequence objs: A list of :class:`AsymUnit` and/or
              :class:`Entity` objects.
       :param str details: Additional text describing this feature.
    """

    type = 'ligand'

    def __init__(self, objs, details=None):
        self.objs, self.details = objs, details
        _ = self._get_entity_type()

    def _all_entities_or_asyms(self):
        return self.objs

    def _get_entity_type(self):
        def _get_entity(x):
            return x if isinstance(x, ihm.Entity) else x.entity
        if any(_get_entity(r).is_polymeric() for r in self.objs):
            raise ValueError(
                "%s can only select non-polymeric entities" % self)
        else:
            return _get_entity(self.objs[0]).type if self.objs else None


class PseudoSiteFeature(Feature):
    """Selection of a pseudo position in the system.

       :param site: The pseudo site to use for the feature.
       :type site: :class:`PseudoSite`
    """

    type = 'pseudo site'

    def __init__(self, site):
        self.site = site

    def _get_entity_type(self):
        return 'other'


class GeometricRestraint(object):
    """A restraint between part of the system and some part of a
       geometric object. See :class:`CenterGeometricRestraint`,
       :class:`InnerSurfaceGeometricRestraint`,
       :class:`OuterSurfaceGeometricRestraint`.

       :param dataset: Reference to the data from which the restraint is
              derived.
       :type dataset: :class:`~ihm.dataset.Dataset`
       :param geometric_object: The geometric object to restrain against.
       :type geometric_object: :class:`ihm.geometry.GeometricObject`
       :param feature: The part of the system to restrain.
       :type feature: :class:`Feature`
       :param distance: Restraint on the distance.
       :type distance: :class:`DistanceRestraint`
       :param float harmonic_force_constant: Force constant, if applicable.
       :param bool restrain_all: If True, all distances are restrained.
    """
    object_characteristic = 'other'
    assembly = None  # no struct_assembly_id for geometric restraints

    def __init__(self, dataset, geometric_object, feature, distance,
                 harmonic_force_constant=None, restrain_all=None,
                 pseudo1=None, pseudo2=None):
        self.dataset = dataset
        self.geometric_object, self.feature = geometric_object, feature
        self.distance, self.restrain_all = distance, restrain_all
        self.harmonic_force_constant = harmonic_force_constant
    _all_features = property(lambda self: (self.feature,))


class CenterGeometricRestraint(GeometricRestraint):
    """A restraint between part of the system and the center of a
       geometric object. See :class:`GeometricRestraint` for a description
       of the parameters.
    """
    object_characteristic = 'center'


class InnerSurfaceGeometricRestraint(GeometricRestraint):
    """A restraint between part of the system and the inner surface of a
       geometric object. See :class:`GeometricRestraint` for a description
       of the parameters.
    """
    object_characteristic = 'inner surface'


class OuterSurfaceGeometricRestraint(GeometricRestraint):
    """A restraint between part of the system and the outer surface of a
       geometric object. See :class:`GeometricRestraint` for a description
       of the parameters.
    """
    object_characteristic = 'outer surface'


class DerivedDistanceRestraint(object):
    """A restraint between two parts of the system, derived from experimental
       data.

       :param dataset: Reference to the data from which the restraint is
              derived.
       :type dataset: :class:`~ihm.dataset.Dataset`
       :param feature1: The first part of the system to restrain.
       :type feature1: :class:`Feature`
       :param feature2: The second part of the system to restrain.
       :type feature2: :class:`Feature`
       :param distance: Restraint on the distance.
       :type distance: :class:`DistanceRestraint`
       :param float probability: Likelihood that restraint is correct (0. - 1.)
       :param bool restrain_all: If True, all distances are restrained.
       :param float mic_value: Value of the Maximal Information Coefficient
              (MIC) for this interaction, if applicable.
    """
    assembly = None  # no struct_assembly_id for derived distance restraints

    def __init__(self, dataset, feature1, feature2, distance,
                 probability=None, restrain_all=None, mic_value=None):
        self.dataset = dataset
        self.feature1, self.feature2 = feature1, feature2
        self.distance, self.restrain_all = distance, restrain_all
        self.probability = probability
        self.mic_value = mic_value
    _all_features = property(lambda self: (self.feature1, self.feature2))


class PredictedContactRestraint(object):
    """A predicted contact between two parts of the system, derived from
       various computational tools.

       :param dataset: Reference to the data from which the restraint is
              derived.
       :type dataset: :class:`~ihm.dataset.Dataset`
       :param resatom1: The first residue or atom to restrain.
       :type resatom1: :class:`ihm.Residue` or :class:`ihm.Atom`
       :param resatom2: The second residue or atom to restrain.
       :type resatom2: :class:`ihm.Residue` or :class:`ihm.Atom`
       :param distance: Restraint on the distance.
       :type distance: :class:`DistanceRestraint`
       :param bool by_residue: If True, the restraint is applied to specific
              residues; otherwise, it is applied to the closest primitive
              object with the highest resolution.
       :param float probability: Likelihood that restraint is correct (0. - 1.)
       :param software: The software used to generate the contact.
       :type software: :class:`~ihm.Software`

    """
    assembly = None  # no struct_assembly_id for predicted contact restraints

    def __init__(self, dataset, resatom1, resatom2, distance,
                 by_residue, probability=None, software=None):
        self.dataset = dataset
        self.resatom1, self.resatom2 = resatom1, resatom2
        self.distance, self.by_residue = distance, by_residue
        self.probability, self.software = probability, software



# coding=utf-8

"""Classes to handle fluorescence data.
   The classes roughly correspond to categories in the
   `FLR dictionary <https://github.com/ihmwg/FLR-dictionary/>`_.
   See the top level :class:`FLRData` class for more information.
"""


class Probe(object):
    """Defines a fluorescent probe.

       This class is not in the FLR dictionary, but it collects all the
       information connected by the probe_ids.

       :param probe_list_entry: A probe list object.
       :type probe_list_entry: :class:`ProbeList`
       :param probe_descriptor: A probe descriptor.
       :type probe_descriptor: :class:`ProbeDescriptor`
    """

    def __init__(self, probe_list_entry=None, probe_descriptor=None):
        self.probe_list_entry = probe_list_entry
        self.probe_descriptor = probe_descriptor

    def __eq__(self, other):
        return self.__dict__ == other.__dict__


class ProbeDescriptor(object):
    """Collects the chemical descriptors for a fluorescent probe.

       This includes the chemical descriptor of the reactive probe and
       the chromophore.

       :param reactive_probe_chem_descriptor: The chemical descriptor for
              the reactive probe.
       :type reactive_probe_chem_descriptor: :class:`ihm.ChemDescriptor`
       :param chromophore_chem_descriptor: The chemical descriptor of the
              chromophore.
       :type chromophore_chem_descriptor: :class:`ihm.ChemDescriptor`
       :param chromophore_center_atom: The atom describing the center
              of the chromophore.
    """

    def __init__(self, reactive_probe_chem_descriptor,
                 chromophore_chem_descriptor, chromophore_center_atom=None):
        self.reactive_probe_chem_descriptor = reactive_probe_chem_descriptor
        self.chromophore_chem_descriptor = chromophore_chem_descriptor
        self.chromophore_center_atom = chromophore_center_atom

    def __eq__(self, other):
        return self.__dict__ == other.__dict__


class ProbeList(object):
    """Store the chromophore name, whether there is a reactive probe
       available, the origin of the probe and the type of linkage of the probe.

       :param str chromophore_name: The name of the chromophore.
       :param bool reactive_probe_flag: Flag to indicate whether a reactive
              probe is given.
       :param str reactive_probe_name: The name of the reactive probe.
       :param str probe_origin: The origin of the probe (intrinsic
              or extrinsic).
       :param str probe_link_type: The type of linkage for the probe (covalent
              or ligand).
    """

    def __init__(self, chromophore_name, reactive_probe_flag=False,
                 reactive_probe_name=None, probe_origin=None,
                 probe_link_type=None):
        self.chromophore_name = chromophore_name
        self.reactive_probe_flag = reactive_probe_flag
        self.reactive_probe_name = reactive_probe_name
        self.probe_origin = probe_origin
        self.probe_link_type = probe_link_type

    def __eq__(self, other):
        return self.__dict__ == other.__dict__


class SampleProbeDetails(object):
    """Connects a probe to a sample.

       :param sample: The sample.
       :type sample: :class:`Sample`
       :param probe: A probe that is attached to the sample.
       :type probe: :class:`Probe`
       :param str fluorophore_type: The type of the fluorophore (donor,
              acceptor, or unspecified).
       :param poly_probe_position: The position on the polymer where
              the dye is attached to.
       :type poly_probe_position: :class:`PolyProbePosition`
       :param str description: A description of the sample-probe-connection.
    """

    def __init__(self, sample, probe, fluorophore_type='unspecified',
                 poly_probe_position=None, description=None):
        self.sample = sample
        self.probe = probe
        self.fluorophore_type = fluorophore_type
        self.description = description
        self.poly_probe_position = poly_probe_position

    def __eq__(self, other):
        return self.__dict__ == other.__dict__


class PolyProbeConjugate(object):
    """Describes the conjugate of polymer residue and probe (including
       possible linker)

       :param sample_probe: The :class:`SampleProbeDetails` object to
              which the conjugate is related.
       :type sample_probe: :class:`SampleProbeDetails`
       :param chem_descriptor: The chemical descriptor of the conjugate
              of polymer residue and probe.
       :type chem_descriptor: :class:`ihm.ChemDescriptor`
       :param bool ambiguous_stoichiometry: Flag whether the labeling
              is ambiguous.
       :param float probe_stoichiometry: The stoichiometry of the
              ambiguous labeling.
    """

    def __init__(self, sample_probe, chem_descriptor,
                 ambiguous_stoichiometry=False, probe_stoichiometry=None):
        self.sample_probe = sample_probe
        self.chem_descriptor = chem_descriptor
        self.ambiguous_stoichiometry = ambiguous_stoichiometry
        self.probe_stoichiometry = probe_stoichiometry

    def __eq__(self, other):
        return self.__dict__ == other.__dict__


class PolyProbePosition(object):
    """Describes a position on the polymer used for attaching the probe.

       This class combines Poly_probe_position, Poly_probe_position_modified,
       and Poly_probe_position_mutated from the FLR dictionary.

       :param resatom: The residue or atom that the probe is attached to.
       :type resatom: :class:`ihm.Residue` or :class:`ihm.Atom`
       :param bool mutation_flag: Flag whether the residue was mutated
              (e.g. a Cys mutation).
       :param bool modification_flag: Flag whether the residue was modified
              (e.g. replacement of a residue with a labeled residue in
              case of nucleic acids).
       :param str auth_name: An author-given name for the position.
       :param mutated_chem_comp_id: The chemical component ID of the
              mutated residue.
       :type modified_chem_descriptor: :class:`ihm.ChemComp`
       :param modified_chem_descriptor: The chemical descriptor of the
              modified residue.
       :type modified_chem_descriptor: :class:`ihm.ChemDescriptor`
    """

    def __init__(self, resatom, mutation_flag=False, modification_flag=False,
                 auth_name=None, mutated_chem_comp_id=None,
                 modified_chem_descriptor=None):
        self.resatom = resatom
        self.mutation_flag = mutation_flag
        self.modification_flag = modification_flag
        self.auth_name = auth_name
        if self.mutation_flag:
            self.mutated_chem_comp_id = mutated_chem_comp_id
        if self.modification_flag:
            self.modified_chem_descriptor = modified_chem_descriptor

    def __eq__(self, other):
        return self.__dict__ == other.__dict__


class Sample(object):
    """Sample corresponds to a measurement.

       :param entity_assembly: The assembly of the entities that was measured.
       :type entity_assembly: :class:`EntityAssembly`
       :param int num_of_probes: The number of probes in the sample.
       :param condition: The sample conditions for the Sample.
       :type condition: :class:`SampleCondition`
       :param str description: A description of the sample.
       :param str details: Details about the sample.
       :param solvent_phase: The solvent phase of the sample (liquid,
              vitrified, or other).
    """

    def __init__(self, entity_assembly, num_of_probes, condition,
                 description=None, details=None, solvent_phase=None):
        self.entity_assembly = entity_assembly
        self.num_of_probes = num_of_probes
        self.condition = condition
        self.description = description
        self.details = details
        self.solvent_phase = solvent_phase

    def __eq__(self, other):
        return self.__dict__ == other.__dict__


class EntityAssembly(object):
    """The assembly of the entities that are in the system.

       :param entity: The entity to add.
       :type entity: :class:`Entity`
       :param num_copies: The number of copies for the entity in the assembly.
    """

    def __init__(self, entity=None, num_copies=0):
        self.entity_list = []
        self.num_copies_list = []
        if entity is not None and num_copies != 0:
            self.add_entity(entity, num_copies)

    def add_entity(self, entity, num_copies):
        if num_copies < 0:
            raise ValueError("Number of copies for Entity must be "
                             "larger than zero.")
        self.entity_list.append(entity)
        self.num_copies_list.append(num_copies)

    def __eq__(self, other):
        return self.__dict__ == other.__dict__


class SampleCondition(object):
    """Description of the sample conditions.

       *Currently this is only text, but will be extended in the future.*

       :param str details: Description of the sample conditions.
    """

    def __init__(self, details=None):
        self.details = details

    def __eq__(self, other):
        return self.__dict__ == other.__dict__


class Experiment(object):
    """The Experiment collects combinations of instrument, experimental
       settings and sample.

       :param instrument: The instrument.
       :type instrument: :class:`Instrument`
       :param inst_setting: The instrument setting.
       :type inst_setting: :class:`InstSetting`
       :param exp_condition: The experimental conditions.
       :type exp_condition: :class:`ExpCondition`
       :param sample: The sample.
       :type sample: :class:`Sample`
       :param details: Details on the experiment.
    """

    def __init__(self, instrument=None, inst_setting=None, exp_condition=None,
                 sample=None, details=None):
        """The Experiment object can either be initiated with empty lists,
           or with an entry for each of them. In this way, an experiment
           object is created and filled with one entry.
        """
        self.instrument_list = []
        self.inst_setting_list = []
        self.exp_condition_list = []
        self.sample_list = []
        self.details_list = []
        if (instrument is not None and inst_setting is not None
                and exp_condition is not None and sample is not None):
            self.add_entry(instrument=instrument, inst_setting=inst_setting,
                           exp_condition=exp_condition,
                           sample=sample, details=details)

    def add_entry(self, instrument, inst_setting, exp_condition, sample,
                  details=None):
        """Entries to the experiment object can also be added one by one.
        """
        self.instrument_list.append(instrument)
        self.inst_setting_list.append(inst_setting)
        self.exp_condition_list.append(exp_condition)
        self.sample_list.append(sample)
        self.details_list.append(details)

    def get_entry_by_index(self, index):
        """Returns the combination of :class:`Instrument`, :class:`InstSetting`,
           :class:`ExpCondition`, :class:`Sample`, and details for a given
           index.
        """
        return (self.instrument_list[index],
                self.inst_setting_list[index],
                self.exp_condition_list[index],
                self.sample_list[index],
                self.details_list[index])

    def __eq__(self, other):
        return ((self.instrument_list == other.instrument_list)
                and (self.inst_setting_list == other.inst_setting_list)
                and (self.exp_condition_list == other.exp_condition_list)
                and (self.sample_list == other.sample_list)
                and (self.details_list == other.details_list))

    def contains(self, instrument, inst_setting, exp_condition, sample):
        """Checks whether a combination of :class:`Instrument`,
           :class:`InstSetting`, :class:`ExpCondition`,
           :class:`Sample` is already included in the experiment object.
        """
        # TODO: possibly extend this by the details_list?
        for i in range(len(self.instrument_list)):
            if ((instrument == self.instrument_list[i])
                    and (inst_setting == self.inst_setting_list[i])
                    and (exp_condition == self.exp_condition_list[i])
                    and (sample == self.sample_list[i])):
                return True
        return False


class Instrument(object):
    """Description of the Instrument used for the measurements.

       *Currently this is only text, but will be extended in the future.*

       :param details: Description of the instrument used for the measurements.
    """
    def __init__(self, details=None):
        self.details = details

    def __eq__(self, other):
        return self.__dict__ == other.__dict__


class InstSetting(object):
    """Description of the instrument settings.

       *Currently this is only text, but will be extended in the future.*

       :param str details: Description of the instrument settings used for
              the measurement (e.g. laser power or size of observation
              volume in case of confocal measurements).
    """
    def __init__(self, details=None):
        self.details = details

    def __eq__(self, other):
        return self.__dict__ == other.__dict__


class ExpCondition(object):
    """Description of the experimental conditions.

    * Currently this is only text, but will be extended in the future.*

    :param str details: Description of the experimental conditions (e.g.
        the temperature at which the experiment was carried out).
    """
    def __init__(self, details=None):
        self.details = details

    def __eq__(self, other):
        return self.__dict__ == other.__dict__


class FRETAnalysis(object):
    """An analysis of FRET data that was performed.

       :param experiment: The Experiment object for this FRET analysis.
       :type experiment: :class:`Experiment`
       :param sample_probe_1: The combination of sample and probe for the
              first probe.
       :type sample_probe_1: :class:`SampleProbeDetails`
       :param sample_probe_2: The combination of sample and probe for the
              second probe.
       :type sample_probe_2: :class:`SampleProbeDetails`
       :param forster_radius: The Förster radius object for this FRET analysis.
       :type forster_radius: :class:`FRETForsterRadius`.
       :param str type: The type of the FRET analysis (intensity-based
              or lifetime-based).
       :param calibration_parameters: The calibration parameters used for
              this analysis (only in case of intensity-based analyses).
       :type calibration_parameters: :class:`FRETCalibrationParameters`
       :param lifetime_fit_model: The fit model used in case of
              lifetime-based analyses.
       :type lifetime_fit_model: :class:`LifetimeFitModel`
       :param ref_measurement_group: The group of reference measurements
             in case of lifetime-based analyses.
       :type ref_measurement_group: :class:`LifetimeRefMeasurementGroup`
       :param str method_name: The method used for the analysis.
       :param float chi_square_reduced: The chi-square reduced as a quality
              measure for the fit.
       :param float donor_only_fraction: The donor-only fraction.
       :param dataset: The dataset used.
       :type dataset: :class:`ihm.dataset.Dataset`
       :param external_file: The external file that contains (results of)
              the analysis.
       :param software: The software used for the analysis.
       :type software: :class:`ihm.Software`
    """

    def __init__(self, experiment, sample_probe_1, sample_probe_2,
                 forster_radius, type, calibration_parameters=None,
                 lifetime_fit_model=None,
                 ref_measurement_group=None,
                 method_name=None, details=None,
                 chi_square_reduced=None, donor_only_fraction=None,
                 dataset=None, external_file=None, software=None):
        if type not in ['lifetime-based', 'intensity-based', None]:
            raise ValueError(
                'FRETAnalysis.type can be \'lifetime-based\' or '
                '\'intensity-based\'. The value is %s' % type)
        self.experiment = experiment
        self.sample_probe_1 = sample_probe_1
        self.sample_probe_2 = sample_probe_2
        self.forster_radius = forster_radius
        self.type = type
        self.calibration_parameters = calibration_parameters
        self.lifetime_fit_model = lifetime_fit_model
        self.ref_measurement_group = ref_measurement_group
        self.method_name = method_name
        self.details = details
        self.chi_square_reduced = chi_square_reduced
        self.donor_only_fraction = donor_only_fraction
        self.dataset = dataset
        self.external_file = external_file
        self.software = software

    def __eq__(self, other):
        return self.__dict__ == other.__dict__


class LifetimeFitModel(object):
    """A lifetime-fit model used for lifetime-based analysis.

        :param str name: The name of the fit model.
        :param str description: A description of the fit model.
        :param external_file: An external file that contains additional
               information on the fit model.
        :param citation: A citation for the fit model.
        :type citation: :class:`ihm.Citation`
    """
    def __init__(self, name, description, external_file=None, citation=None):
        self.name = name
        self.description = description
        self.external_file = external_file
        self.citation = citation

    def __eq__(self, other):
        return self.__dict__ == other.__dict__


class RefMeasurementGroup(object):
    """A Group containing reference measurements for lifetime-based analysis.

        :param str details: Details on the Group of reference measurements.
    """
    def __init__(self, details=None):
        self.details = details
        self.ref_measurement_list = []
        self.num_measurements = len(self.ref_measurement_list)

    def add_ref_measurement(self, ref_measurement):
        """Add a lifetime reference measurement to a ref_measurement_group."""
        self.ref_measurement_list.append(ref_measurement)
        self.num_measurements = len(self.ref_measurement_list)

    def get_info(self):
        return self.ref_measurement_list

    def __eq__(self, other):
        return self.__dict__ == other.__dict__


class RefMeasurement(object):
    """A reference measurement for lifetime-based analysis.

        :param ref_sample_probe: The combination of sample and probe used
               for the reference measurement.
        :type ref_sample_probe: :class:`SampleProbeDetails`
        :param str details: Details on the measurement.
        :param list_of_lifetimes: A list of the results from the reference
               measurement.
        :type list_of_lifetimes: List of :class:`RefMeasurementLifetime`
    """
    def __init__(self, ref_sample_probe, details=None, list_of_lifetimes=None):
        self.ref_sample_probe = ref_sample_probe
        self.details = details
        self.list_of_lifetimes = \
            list_of_lifetimes if list_of_lifetimes is not None else []
        self.num_species = len(self.list_of_lifetimes)

    def add_lifetime(self, lifetime):
        """Add a lifetime to the list_of_lifetimes."""
        self.list_of_lifetimes.append(lifetime)
        self.num_species = len(self.list_of_lifetimes)

    def __eq__(self, other):
        return self.__dict__ == other.__dict__


class RefMeasurementLifetime(object):
    """Lifetime for a species in a reference measurement.
        :param float species_fraction: The species-fraction for the
               respective lifetime.
        :param float lifetime: The lifetime (in ns).
        :param str species_name: A name for the species.
    """
    def __init__(self, species_fraction, lifetime, species_name=None):
        self.species_fraction = species_fraction
        self.lifetime = lifetime
        self.species_name = species_name

    def __eq__(self, other):
        return self.__dict__ == other.__dict__


class FRETDistanceRestraintGroup(object):
    """A collection of FRET distance restraints that are used together.
    """
    def __init__(self):
        self.distance_restraint_list = []

    def add_distance_restraint(self, distance_restraint):
        """Add a distance restraint to a distance_restraint_group"""
        self.distance_restraint_list.append(distance_restraint)

    def get_info(self):
        return self.distance_restraint_list

    def __eq__(self, other):
        return self.__dict__ == other.__dict__


class FRETDistanceRestraint(object):
    """A distance restraint from FRET.

       :param sample_probe_1: The combination of sample and probe for
              the first probe.
       :type sample_probe_1: :class:`SampleProbeDetails`
       :param sample_probe_2: The combination of sample and probe for
              the second probe.
       :type sample_probe_2: :class:`SampleProbeDetails`
       :param analysis: The FRET analysis from which the distance
              restraint originated.
       :type analysis: :class:`FRETAnalysis`
       :param float distance: The distance of the restraint.
       :param float distance_error_plus: The (absolute, e.g. in Angstrom) error
              in the upper direction, such that
              ``upper boundary = distance + distance_error_plus``.
       :param float distance_error_minus: The (absolute, e.g. in Angstrom)
              error in the lower direction, such that
              ``lower boundary = distance + distance_error_minus``.
       :param str distance_type: The type of distance (<R_DA>, <R_DA>_E,
              or R_mp).
       :param state: The state the distance restraints is connected to.
              Important for multi-state models.
       :type state: :class:`ihm.model.State`
       :param float population_fraction: The population fraction of the state
              in case of multi-state models.
       :param peak_assignment: The method how a peak was assigned.
       :type peak_assignment: :class:`PeakAssignment`
    """

    def __init__(self, sample_probe_1, sample_probe_2, analysis, distance,
                 distance_error_plus=0., distance_error_minus=0.,
                 distance_type=None, state=None, population_fraction=0.,
                 peak_assignment=None):
        self.sample_probe_1 = sample_probe_1
        self.sample_probe_2 = sample_probe_2
        self.state = state
        self.analysis = analysis
        self.distance = distance
        self.distance_error_plus = distance_error_plus
        self.distance_error_minus = distance_error_minus
        self.distance_type = distance_type
        self.population_fraction = population_fraction
        self.peak_assignment = peak_assignment

    def __eq__(self, other):
        return self.__dict__ == other.__dict__


class FRETForsterRadius(object):
    """The FRET Förster radius between two probes.

       :param donor_probe: The donor probe.
       :type donor_probe: :class:`Probe`
       :param acceptor_probe: The acceptor probe.
       :type acceptor_probe: :class:`Probe`
       :param float forster_radius: The Förster radius between the two probes.
       :param float reduced_forster_radius: The reduced Förster radius between
              the two probes.
    """

    def __init__(self, donor_probe, acceptor_probe, forster_radius,
                 reduced_forster_radius=None):
        self.donor_probe = donor_probe
        self.acceptor_probe = acceptor_probe
        self.forster_radius = forster_radius
        self.reduced_forster_radius = reduced_forster_radius

    def __eq__(self, other):
        return self.__dict__ == other.__dict__


class FRETCalibrationParameters(object):
    """The calibration parameter from the FRET measurements.
        For the definitions of the parameters see
        Hellenkamp et al. Nat. Methods 2018.

        :param float phi_acceptor: The quantum yield of the acceptor.
        :param float alpha: The alpha parameter.
        :param float alpha_sd: The standard deviation of the alpha parameter.
        :param float gg_gr_ratio: The ratio of the green and red detection
               efficiencies.
        :param float beta: The beta parameter.
        :param float gamma: The gamma parameter.
        :param float delta: The delta parameter.
        :param float a_b: The fraction of bright molecules.
    """

    def __init__(self, phi_acceptor=None, alpha=None, alpha_sd=None,
                 gg_gr_ratio=None, beta=None, gamma=None, delta=None,
                 a_b=None):
        self.phi_acceptor = phi_acceptor
        self.alpha = alpha
        self.alpha_sd = alpha_sd
        self.gg_gr_ratio = gg_gr_ratio
        self.beta = beta
        self.gamma = gamma
        self.delta = delta
        self.a_b = a_b

    def __eq__(self, other):
        return self.__dict__ == other.__dict__


class PeakAssignment(object):
    """The method of peak assignment in case of multiple peaks,
        e.g. by population.

        :param str method_name: The method used for peak assignment.
        :param str details: The details of the peak assignment procedure.
    """
    def __init__(self, method_name, details=None):
        self.method_name = method_name
        self.details = details

    def __eq__(self, other):
        return self.__dict__ == other.__dict__


class FRETModelQuality(object):
    """The quality measure for a Model based on FRET data.

       :param model: The model being described.
       :type model: :class:`ihm.model.Model`
       :param chi_square_reduced: The quality of the model in terms of
              chi_square_reduced based on the Distance restraints used
              for the modeling.
       :param dataset_group: The group of datasets that was used for the
              quality estimation.
       :type dataset_group: :class:`ihm.dataset.DatasetGroup`
       :param method: The method used for judging the model quality.
       :param str details: Details on the model quality.
    """

    def __init__(self, model, chi_square_reduced, dataset_group,
                 method, details=None):
        self.model = model
        self.chi_square_reduced = chi_square_reduced
        self.dataset_group = dataset_group
        self.method = method
        self.details = details

    def __eq__(self, other):
        return self.__dict__ == other.__dict__


class FRETModelDistance(object):
    """The distance in a model for a certain distance restraint.

       :param restraint: The Distance restraint.
       :type restraint: :class:`FRETDistanceRestraint`
       :param model: The model the distance applies to.
       :type model: :class:`ihm.model.Model`
       :param distance: The distance obtained for the distance restraint
              in the current model.
       :param distance_deviation: The deviation of the distance in the
              model compared to the value of the distance restraint.
    """

    def __init__(self, restraint, model, distance,
                 distance_deviation=None):
        self.restraint = restraint
        self.model = model
        self.distance = distance
        self.distance_deviation = distance_deviation
        if self.distance_deviation is None and self.restraint is not None:
            self.calculate_deviation()

    def calculate_deviation(self):
        if self.distance_deviation is None and self.restraint is not None:
            self.distance_deviation = \
                float(self.restraint.distance) - float(self.distance)

    def update_deviation(self):
        if self.restraint is not None:
            self.distance_deviation = \
                float(self.restraint.distance) - float(self.distance)

    def __eq__(self, other):
        return self.__dict__ == other.__dict__


class FPSModeling(object):
    """Collect the modeling parameters for different steps of FPS,
       e.g. Docking, Refinement, or Error estimation.

       :param protocol: The modeling protocol to which the FPS modeling
              step belongs.
       :type protocol: :class:`ihm.protocol.Protocol`
       :param restraint_group: The restraint group used for the modeling.
       :type restraint_group: :class:`FRETDistanceRestraintGroup`
       :param global_parameter: The global FPS parameters used.
       :type global_parameter: :class:`FPSGlobalParameters`
       :param str probe_modeling_method: either "AV" or "MPP".
       :param str details: Details on the FPS modeling.
    """

    def __init__(self, protocol, restraint_group,
                 global_parameter, probe_modeling_method, details=None):
        self.protocol = protocol
        self.restraint_group = restraint_group
        self.global_parameter = global_parameter
        self.probe_modeling_method = probe_modeling_method
        self.details = details

    def __eq__(self, other):
        return self.__dict__ == other.__dict__


class FPSGlobalParameters(object):
    """The global parameters in the FPS program.

       *For a description of the parameters, see also the FPS manual.*

       :param float forster_radius: The Förster radius used in the FPS program.
       :param int conversion_function_polynom_order: Order of the polynom for
              the conversion function between Rmp and <RDA>E.
       :param int repetition: The number of repetitions.
       :param float av_grid_rel: The AV grid spacing relative to the smallest
              dye or linker dimension.
       :param float av_min_grid_a: The minimal AV grid spacing in Angstrom.
       :param float av_allowed_sphere: The allowed sphere radius.
       :param int av_search_nodes: Number of neighboring positions to be
              scanned for clashes.
       :param float av_e_samples_k: The number of samples for calculation
              of E (in thousand).
       :param float sim_viscosity_adjustment: Daming rate during docking
              and refinement.
       :param float sim_dt_adjustment: Time step during simulation.
       :param float sim_max_iter_k: Maximal number of iterations (in thousand).
       :param float sim_max_force: Maximal force.
       :param float sim_clash_tolerance_a: Clash tolerance in Angstrom.
       :param float sim_reciprocal_kt: reciprocal kT.
       :param str sim_clash_potential: The clash potential.
       :param float convergence_e: Convergence criterion E.
       :param float convergence_k: Convergence criterion K.
       :param float convergence_f: Convergence criterion F.
       :param float convergence_t: Convergence criterion T.
       :param str optimized_distances: Which distances are optimized?

    """
    def __init__(self, forster_radius, conversion_function_polynom_order,
                 repetition, av_grid_rel, av_min_grid_a, av_allowed_sphere,
                 av_search_nodes, av_e_samples_k, sim_viscosity_adjustment,
                 sim_dt_adjustment, sim_max_iter_k, sim_max_force,
                 sim_clash_tolerance_a, sim_reciprocal_kt, sim_clash_potential,
                 convergence_e, convergence_k, convergence_f, convergence_t,
                 optimized_distances='All'):
        self.forster_radius = forster_radius
        self.conversion_function_polynom_order \
            = conversion_function_polynom_order
        self.repetition = repetition
        self.av_grid_rel = av_grid_rel
        self.av_min_grid_a = av_min_grid_a
        self.av_allowed_sphere = av_allowed_sphere
        self.av_search_nodes = av_search_nodes
        self.av_e_samples_k = av_e_samples_k
        self.sim_viscosity_adjustment = sim_viscosity_adjustment
        self.sim_dt_adjustment = sim_dt_adjustment
        self.sim_max_iter_k = sim_max_iter_k
        self.sim_max_force = sim_max_force
        self.sim_clash_tolerance_a = sim_clash_tolerance_a
        self.sim_reciprocal_kt = sim_reciprocal_kt
        self.sim_clash_potential = sim_clash_potential
        self.convergence_e = convergence_e
        self.convergence_k = convergence_k
        self.convergence_f = convergence_f
        self.convergence_t = convergence_t
        self.optimized_distances = optimized_distances

    def __eq__(self, other):
        return self.__dict__ == other.__dict__


class FPSAVModeling(object):
    """FPS modeling using AV.
       This object connects the FPS_modeling step, the sample_probe and
       the respective AV parameters.

       :param fps_modeling: The FPS modeling ID.
       :type fps_modeling: :class:`FPSModeling`
       :param sample_probe: The Sample probe ID.
       :type sample_probe: :class:`SampleProbeDetails`
       :param parameter: The FPS AV parameters used.
       :type parameter: :class:`FPSAVParameter`
    """

    def __init__(self, fps_modeling, sample_probe, parameter):
        # fps_modeling is the object containing information on the
        # ihm modeling protocol, the restraint group and the global
        # FPS parameters
        self.fps_modeling = fps_modeling
        self.sample_probe = sample_probe
        self.parameter = parameter

    def __eq__(self, other):
        return self.__dict__ == other.__dict__


class FPSAVParameter(object):
    """The AV parameters used for the modeling using FPS.

       :param int num_linker_atoms: The number of atoms in the linker.
       :param float linker_length: The length of the linker in Angstrom.
       :param float linker_width: The width of the linker in Angstrom.
       :param float probe_radius_1: The first radius of the probe.
       :param float probe_radius_2: If AV3 is used, the second radius
              of the probe.
       :param float probe_radius_3: If AV3 is used, the third radius
              of the probe.
    """

    def __init__(self, num_linker_atoms, linker_length, linker_width,
                 probe_radius_1, probe_radius_2=None, probe_radius_3=None):
        self.num_linker_atoms = num_linker_atoms
        self.linker_length = linker_length
        self.linker_width = linker_width
        self.probe_radius_1 = probe_radius_1
        self.probe_radius_2 = probe_radius_2
        self.probe_radius_3 = probe_radius_3

    def __eq__(self, other):
        return self.__dict__ == other.__dict__


class FPSMPPModeling(object):
    """Maps the FPSModeling object to a mean probe position and connects it
       to the reference coordinate system.

       :param fps_modeling: The FPS modeling object.
       :type fps_modeling: :class:`FPSModeling`
       :param mpp: The ID of the mean probe position.
       :type mpp: :class:`FPSMeanProbePosition`
       :param mpp_atom_position_group:
       :type mpp_atom_position_group: :class:`FPSMPPAtomPositionGroup`
    """

    def __init__(self, fps_modeling, mpp, mpp_atom_position_group):
        # fps_modeling is the object containing information on the
        # ihm modeling protocol, the restraint group and the global
        # FPS parameters
        self.fps_modeling = fps_modeling
        self.mpp = mpp
        self.mpp_atom_position_group = mpp_atom_position_group

    def __eq__(self, other):
        return self.__dict__ == other.__dict__


class FPSMeanProbePosition(object):
    """The mean probe position of an AV, which can be used instead of an AV.

       *It is usually not recommended to use this. Use AVs instead.*
       The coordinates are with respect to a reference coordinate system
       defined by :class:`FPSMPPAtomPositionGroup`.

       :param sample_probe: The Sample probe.
       :type sample_probe: :class:`SampleProbeDetails`
       :param float x: The x-coordinate of the mean probe position.
       :param float y: The y-coordinate of the mean probe position.
       :param float z: The z-coordinate of the mean probe position.
    """

    def __init__(self, sample_probe, x, y, z):
        self.sample_probe = sample_probe
        self.x, self.y, self.z = x, y, z

    def __eq__(self, other):
        return self.__dict__ == other.__dict__


class FPSMPPAtomPositionGroup(object):
    """A group of atom positions used to define the coordinate system
       of a mean probe position.
       *Not part of the FLR dictionary.*
    """
    def __init__(self):
        self.mpp_atom_position_list = []

    def add_atom_position(self, atom_position):
        self.mpp_atom_position_list.append(atom_position)

    def __eq__(self, other):
        return self.__dict__ == other.__dict__


class FPSMPPAtomPosition(object):
    """An atom used to describe the coordinate system for a mean probe position

       :param atom: The atom being described.
       :type atom: :class:`ihm.Atom`
       :param float x: The x-coordinate of the atom.
       :param float y: The y-coordinate of the atom.
       :param float z: The z-coordinate of the atom.
    """

    # atoms describing the coordinate system for a mean probe position
    def __init__(self, atom, x, y, z):
        self.atom, self.x, self.y, self.z = atom, x, y, z

    def __eq__(self, other):
        return self.__dict__ == other.__dict__


class FLRData(object):
    """A collection of the fluorescence data to be added to the system.

       Instances of this class are generally added to
       :attr:`~ihm.System.flr_data`.
    """
    def __init__(self):
        #: All groups of FRET distance restraints.
        #: See :class:`FRETDistanceRestraintGroup`.
        self.distance_restraint_groups = []

        #: All conjugates of polymer residue and probe.
        #: See :class:`PolyProbeConjugate`.
        self.poly_probe_conjugates = []

        #: All quality measures for models based on FRET data.
        #: See :class:`FRETModelQuality`.
        self.fret_model_qualities = []

        #: All distances in models for distance restraints.
        #: See :class:`FRETModelDistance`.
        self.fret_model_distances = []

        #: All modeling objects.
        #: See :class:`FPSAVModeling` and :class:`FPSMPPModeling`.
        self.fps_modeling = []

        # The following dictionaries are so far only used when reading data
        self._collection_flr_experiment = {}
        self._collection_flr_inst_setting = {}
        self._collection_flr_exp_condition = {}
        self._collection_flr_instrument = {}
        self._collection_flr_entity_assembly = {}
        self._collection_flr_sample_condition = {}
        self._collection_flr_sample = {}
        self._collection_flr_sample_probe_details = {}
        self._collection_flr_probe = {}
        self._collection_flr_poly_probe_position = {}
        self._collection_flr_poly_probe_position_modified = {}
        self._collection_flr_poly_probe_position_mutated = {}
        self._collection_flr_poly_probe_conjugate = {}
        self._collection_flr_fret_forster_radius = {}
        self._collection_flr_fret_calibration_parameters = {}
        self._collection_flr_fret_analysis = {}
        self._collection_flr_lifetime_fit_model = {}
        self._collection_flr_ref_measurement_group = {}
        self._collection_flr_ref_measurement = {}
        self._collection_flr_ref_measurement_lifetime = {}
        self._collection_flr_peak_assignment = {}
        self._collection_flr_fret_distance_restraint = {}
        self._collection_flr_fret_distance_restraint_group = {}
        self._collection_flr_fret_model_quality = {}
        self._collection_flr_fret_model_distance = {}
        self._collection_flr_fps_global_parameters = {}
        self._collection_flr_fps_modeling = {}
        self._collection_flr_fps_av_parameter = {}
        self._collection_flr_fps_av_modeling = {}
        self._collection_flr_fps_mean_probe_position = {}
        self._collection_flr_fps_mpp_atom_position = {}
        self._collection_flr_fps_mpp_modeling = {}

    def _all_distance_restraints(self):
        """Yield all FRETDistanceRestraint objects"""
        for rg in self.distance_restraint_groups:
            for r in rg.distance_restraint_list:
                yield r

    def _all_analyses(self):
        """Yield all FRETAnalysis objects"""
        for r in self._all_distance_restraints():
            yield r.analysis

    def _all_peak_assignments(self):
        """Yield all PeakAssignment objects"""
        for r in self._all_distance_restraints():
            yield r.peak_assignment

    def _all_experiments(self):
        """Yield all Experiment objects"""
        for a in self._all_analyses():
            yield a.experiment

    def _all_forster_radii(self):
        """Yield all FRETForsterRadius objects"""
        for a in self._all_analyses():
            yield a.forster_radius

    def _all_calibration_parameters(self):
        """Yield all FRETCalibrationParameters objects"""
        for a in self._all_analyses():
            if a.type == 'intensity-based':
                yield a.calibration_parameters

    def _all_lifetime_fit_models(self):
        """Yield all LifetimeFitModel objects"""
        for a in self._all_analyses():
            if a.type == 'lifetime-based':
                yield a.lifetime_fit_model

    def _all_ref_measurement_groups(self):
        """Yield all RefMeasurementGroup objects"""
        for a in self._all_analyses():
            if a.type == 'lifetime-based':
                yield a.ref_measurement_group

    def _all_ref_measurements(self):
        """Yield all RefMeasurement objects"""
        for rg in self._all_ref_measurement_groups():
            for x in rg.ref_measurement_list:
                yield x

    def _all_ref_measurement_lifetimes(self):
        """Yield all RefMeasurementLifetime objects"""
        for r in self._all_ref_measurements():
            for x in r.list_of_lifetimes:
                yield x

    def _all_sample_probe_details(self):
        """Yield all SampleProbeDetails objects"""
        for r in self._all_distance_restraints():
            yield r.sample_probe_1
            yield r.sample_probe_2
        for r in self._all_ref_measurements():
            yield r.ref_sample_probe

    def _all_samples(self):
        """Yield all Sample objects"""
        for s in self._all_sample_probe_details():
            yield s.sample

    def _all_probes(self):
        """Yield all Probe objects"""
        for s in self._all_sample_probe_details():
            yield s.probe

    def _all_poly_probe_positions(self):
        """Yield all PolyProbePosition objects"""
        for s in self._all_sample_probe_details():
            yield s.poly_probe_position

    def _all_inst_settings(self):
        """Yield all InstSetting objects"""
        for e in self._all_experiments():
            for s in e.inst_setting_list:
                yield s

    def _all_exp_conditions(self):
        """Yield all ExpCondition objects"""
        for e in self._all_experiments():
            for s in e.exp_condition_list:
                yield s

    def _all_instruments(self):
        """Yield all Instrument objects"""
        for e in self._all_experiments():
            for s in e.instrument_list:
                yield s

    def _all_fps_modeling(self):
        """Yield all FPSModeling objects"""
        for m in self.fps_modeling:
            yield m.fps_modeling

    def _all_fps_global_parameters(self):
        """Yield all FPSGlobalParameters objects"""
        for m in self._all_fps_modeling():
            yield m.global_parameter

    def _all_fps_av_modeling(self):
        """Yield all FPSAVModeling objects"""
        for m in self.fps_modeling:
            if isinstance(m, FPSAVModeling):
                yield m

    def _all_fps_av_parameter(self):
        """Yield all FPSAVParameter objects"""
        for m in self._all_fps_av_modeling():
            yield m.parameter

    def _all_fps_mpp_modeling(self):
        """Yield all FPSMPPModeling objects"""
        for m in self.fps_modeling:
            if isinstance(m, FPSMPPModeling):
                yield m

    def _all_fps_mean_probe_position(self):
        """Yield all FPSMeanProbePosition objects"""
        for m in self._all_fps_mpp_modeling():
            yield m.mpp

    def _all_fps_atom_position_group(self):
        """Yield all FPSMPPAtomPositionGroup objects"""
        for m in self._all_fps_mpp_modeling():
            yield m.mpp_atom_position_group

    def _all_flr_chemical_descriptors(self):
        """Collect the chemical descriptors from the flr part.
           *This might contain duplicates.*
        """
        # collect from all distance_restraint_groups
        for drgroup in self.distance_restraint_groups:
            # collect from all distance restraints
            for dr in drgroup.distance_restraint_list:
                # collect from both sample_probe_1 and sample_probe_2
                for this_sample_probe in (dr.sample_probe_1,
                                          dr.sample_probe_2):
                    # collect from the probe
                    probe = this_sample_probe.probe
                    # reactive probe
                    yield probe.probe_descriptor.reactive_probe_chem_descriptor
                    # chromophore
                    yield probe.probe_descriptor.chromophore_chem_descriptor
                    # collect from the poly_probe_position
                    pos = this_sample_probe.poly_probe_position
                    # modified chem descriptor
                    if pos.modification_flag:
                        yield pos.modified_chem_descriptor
                # collect from all analyses if they are lifetime-based
                a = dr.analysis
                if a.type == 'lifetime-based':
                    # RefMeasurementGroup
                    rg = a.ref_measurement_group
                    # collect from all RefMeasurements
                    for rm in rg.ref_measurement_list:
                        # collect from the ref_sample_probe
                        this_ref_sample_probe = rm.ref_sample_probe
                        probe = this_ref_sample_probe.probe
                        pd = probe.probe_descriptor
                        # reactive probe
                        yield pd.reactive_probe_chem_descriptor
                        # chromophore
                        yield pd.chromophore_chem_descriptor
                        # collect from the poly_probe_position
                        pos = this_ref_sample_probe.poly_probe_position
                        # modified chem descriptor
                        if pos.modification_flag:
                            yield pos.modified_chem_descriptor
        # and collect from all poly_probe_conjugates
        for c in self.poly_probe_conjugates:
            yield c.chem_descriptor



"""Classes for handling the analysis of a modeling run.
"""


class Step(object):
    """A single step in an :class:`Analysis`.

       Normally one of the base classes is used; see :class:`FilterStep`,
       :class:`ClusterStep`, :class:`RescoreStep`, :class:`ValidationStep`,
       :class:`OtherStep`, and :class:`EmptyStep`.

       :param str feature: feature energy/score;RMSD;dRMSD;other
       :param int num_models_begin: The number of models at the beginning
              of the step
       :param int num_models_end: The number of models at the end of the step
       :param assembly: The part of the system analyzed in this step
       :type assembly: :class:`~ihm.Assembly`
       :param dataset_group: The collection of datasets used in this analysis,
              if applicable
       :type dataset_group: :class:`~ihm.dataset.DatasetGroup`
       :param software: The software used in this step
       :type software: :class:`~ihm.Software`
       :param script_file: Reference to the external file containing the
              script used in this step (usually a
              :class:`~ihm.location.WorkflowFileLocation`).
       :type script_file: :class:`~ihm.location.Location`
       :param str details: Additional text describing this step
    """
    type = 'other'

    def __init__(self, feature, num_models_begin, num_models_end,
                 assembly=None, dataset_group=None, software=None,
                 script_file=None, details=None):
        self.assembly, self.dataset_group = assembly, dataset_group
        self.feature, self.software = feature, software
        self.num_models_begin = num_models_begin
        self.num_models_end = num_models_end
        self.script_file = script_file
        self.details = details


class FilterStep(Step):
    """A single filtering step in an :class:`Analysis`.
       See :class:`Step` for a description of the parameters.
    """
    type = 'filter'


class ClusterStep(Step):
    """A single clustering step in an :class:`Analysis`.
       See :class:`Step` for a description of the parameters.
    """
    type = 'cluster'


class RescoreStep(Step):
    """A single rescoring step in an :class:`Analysis`.
       See :class:`Step` for a description of the parameters.
    """
    type = 'rescore'


class ValidationStep(Step):
    """A single validation step in an :class:`Analysis`.
       See :class:`Step` for a description of the parameters.
    """
    type = 'validation'


class EmptyStep(Step):
    """A 'do nothing' step in an :class:`Analysis`.
       This can be used if modeling outputs were used directly without
       any kind of analysis.
    """
    type = 'none'

    def __init__(self):
        super(EmptyStep, self).__init__(
            feature='none', num_models_begin=None, num_models_end=None)


class Analysis(object):
    """Analysis of a modeling run.
       Each analysis consists of a number of steps (e.g. filtering,
       rescoring, clustering). A modeling run may be followed by any number
       of separate analyses.

       See :attr:`ihm.protocol.Protocol.analyses`.
    """
    def __init__(self):
        #: All analysis steps (:class:`Step` objects)
        self.steps = []



"""Classes for handling modeling protocols.
"""


class Step(object):
    """A single step in a :class:`Protocol`.

       :param assembly: The part of the system modeled in this step
       :type assembly: :class:`~ihm.Assembly`
       :param dataset_group: The collection of datasets used in this modeling
       :type dataset_group: :class:`~ihm.dataset.DatasetGroup`
       :param str method: Description of the method used (e.g. "Monte Carlo")
       :param str name: A descriptive name for the step
       :param int num_models_begin: The number of models at the beginning
              of the step
       :param int num_models_end: The number of models at the end of the step
       :param software: The software used in this step
       :type software: :class:`~ihm.Software`
       :param script_file: Reference to the external file containing the
              script used in this step (usually a
              :class:`~ihm.location.WorkflowFileLocation`).
       :type script_file: :class:`~ihm.location.Location`
       :param bool multi_scale: Indicates if the modeling is multi-scale
       :param bool multi_state: Indicates if the modeling is multi-state
       :param bool ordered: Indicates if the modeling is ordered
       :param str description: Additional text describing the step
    """
    def __init__(self, assembly, dataset_group, method, num_models_begin=None,
                 num_models_end=None, software=None, script_file=None,
                 multi_scale=False, multi_state=False, ordered=False,
                 name=None, description=None):
        self.assembly = assembly
        self.dataset_group = dataset_group
        self.method = method
        self.num_models_begin = num_models_begin
        self.num_models_end = num_models_end
        self.multi_scale, self.multi_state = multi_scale, multi_state
        self.software, self.ordered, self.name = software, ordered, name
        self.script_file = script_file
        self.description = description


class Protocol(object):
    """A modeling protocol.
       Each protocol consists of a number of protocol steps (e.g. sampling,
       refinement) followed by a number of analyses.

       Normally a protocol is passed to one or more :class:`~ihm.model.Model`
       objects, although unused protocols can still be included in the file
       if desired by adding them to :attr:`~ihm.System.orphan_protocols`.
    """
    def __init__(self, name=None):
        self.name = name

        #: All modeling steps (:class:`Step` objects)
        self.steps = []

        #: All analyses (:class:`~ihm.analysis.Analysis` objects)
        self.analyses = []



"""Classes to extract metadata from various input files.

   Often input files contain metadata that would be useful to include in
   the mmCIF file, but the metadata is stored in a different way for each
   domain-specific file type. For example, MRC files used for electron
   microscopy maps may contain an EMDB identifier, which the mmCIF file
   can point to in preference to the local file.

   This module provides classes for each file type to extract suitable
   metadata where available.
"""

import ihm
from . import location, dataset, startmodel, util
from .startmodel import SequenceIdentityDenominator
import ihm.source
import ihm.citations

import operator
import struct
import json
import warnings
import sys
import re

# Handle different naming of urllib in Python 2/3
try:
    import urllib.request
    import urllib.error
except ImportError:
    class MockUrlLib(object):
        pass
    urllib = MockUrlLib()
    urllib.request = urllib.error = __import__('urllib2')


class Parser(object):
    """Base class for all metadata parsers."""

    def parse_file(self, filename):
        """Extract metadata from the given file.

           :param str filename: the file to extract metadata from.
           :return: a dict with extracted metadata (generally including
                    a :class:`~ihm.dataset.Dataset`)."""
        pass


class MRCParser(Parser):
    """Extract metadata from an EM density map (MRC file)."""

    def parse_file(self, filename):
        """Extract metadata. See :meth:`Parser.parse_file` for details.

           :return: a dict with key `dataset` pointing to the density map,
                    as an EMDB entry if the file contains EMDB headers,
                    otherwise to the file itself.

           If the file turns out to be an EMDB entry, this will also query
           the EMDB web API (if available) to extract version information
           and details for the dataset.
        """
        emdb = self._get_emdb(filename)
        if emdb:
            loc = _ParsedEMDBLocation(emdb)
        else:
            loc = location.InputFileLocation(
                filename, details="Electron microscopy density map")
        return {'dataset': dataset.EMDensityDataset(loc)}

    def _get_emdb(self, filename):
        """Return the EMDB id of the file, or None."""
        r = re.compile(b'EMDATABANK\\.org.*(EMD\\-\\d+)')
        with open(filename, 'rb') as fh:
            fh.seek(220)  # Offset of number of labels
            num_labels_raw = fh.read(4)
            # Number of labels in MRC is usually a very small number, so it's
            # very likely to be the smaller of the big-endian and little-endian
            # interpretations of this field
            num_labels_big, = struct.unpack_from('>i', num_labels_raw)
            num_labels_little, = struct.unpack_from('<i', num_labels_raw)
            num_labels = min(num_labels_big, num_labels_little)
            for i in range(num_labels):
                label = fh.read(80).strip()
                m = r.search(label)
                if m:
                    if sys.version_info[0] < 3:
                        return m.group(1)
                    else:
                        return m.group(1).decode('ascii')


class _ParsedEMDBLocation(location.EMDBLocation):
    """Like an EMDBLocation, but looks up version and details from EMDB
       when they are requested (unless they are set to other values)."""
    def __init__(self, emdb):
        self.__emdb_info = None
        super(_ParsedEMDBLocation, self).__init__(
            db_code=emdb, version=None, details=None)
        self.__emdb_info = None

    def __get_version(self):
        self._get_emdb_info()
        return self.__emdb_info[0]

    def __set_version(self, val):
        if self.__emdb_info is None:
            self.__emdb_info = [None, None]
        self.__emdb_info[0] = val

    def __get_details(self):
        self._get_emdb_info()
        return self.__emdb_info[1] or "Electron microscopy density map"

    def __set_details(self, val):
        if self.__emdb_info is None:
            self.__emdb_info = [None, None]
        self.__emdb_info[1] = val

    def _get_emdb_info(self):
        """Query EMDB API and get version & details of a given entry"""
        if self.__emdb_info is not None:
            return
        req = urllib.request.Request(
            'https://www.ebi.ac.uk/pdbe/api/emdb/entry/summary/%s'
            % self.access_code, None, {})
        try:
            response = urllib.request.urlopen(req, timeout=10)
        except urllib.error.URLError as err:
            warnings.warn("EMDB API query failed; using default metadata "
                          "for MRC file; %s" % str(err))
            self.__emdb_info = [None, None]
            return
        contents = json.load(response)
        keys = list(contents.keys())
        info = contents[keys[0]][0]['deposition']
        # JSON values are always Unicode, but on Python 2 we want non-Unicode
        # strings, so convert to ASCII
        if sys.version_info[0] < 3:
            self.__emdb_info = [info['map_release_date'].encode('ascii'),
                                info['title'].encode('ascii')]
        else:
            self.__emdb_info = [info['map_release_date'], info['title']]

    version = property(__get_version, __set_version)
    details = property(__get_details, __set_details)


def _get_swiss_model_metadata(filename):
    """Extract and return metadata from SWISS-MODEL PDB REMARK headers"""
    meta = {}
    with open(filename) as fh:
        in_header = None
        for line in fh:
            if line.startswith('ATOM'):
                break
            if line.startswith('REMARK   3 '):
                if line.startswith('REMARK   3 MODEL INFORMATION'):
                    in_header = {}
                    meta['info'] = in_header
                elif line.startswith('REMARK   3 TEMPLATE'):
                    in_header = {}
                    meta[line[11:].rstrip('\r\n  ')] = in_header
                elif in_header is not None:
                    linedata = line[11:].rstrip('\r\n ')
                    if linedata:
                        key, val = linedata.split(None, 1)
                        if key == 'ALN':
                            chain, tpltgt, seq = val.split()
                            key = (chain, tpltgt)
                            in_header[key] = in_header.get(key, '') + seq
                        elif key in ('CHAIN', 'MMCIF', 'LIGND'):
                            in_header.setdefault(key, []).append(val)
                        else:
                            in_header[key] = val
    return meta


def _parse_seq(seq):
    """Get a primary sequence and its length (without gaps)"""
    return seq, len(seq.replace('-', ''))


def _get_aligned_region(tgt_seq, tmpl_seq):
    """Given two primary sequences, return the range of each that is
       aligned (i.e. from the first aligned residue in both sequences to
       the last)"""
    first = True
    tgt_pos = 0
    tmpl_pos = 0
    start_align = end_align = None
    for tgt, tmpl in zip(tgt_seq, tmpl_seq):
        if tgt != '-':
            tgt_pos += 1
        if tmpl != '-':
            tmpl_pos += 1
            if tgt != '-':
                end_align = (tgt_pos, tmpl_pos)
                if first:
                    start_align = end_align
                    first = False
    if first:
        raise ValueError("Cannot parse empty alignment")
    return (start_align[0], end_align[0]), (start_align[1], end_align[1])


class PDBParser(Parser):
    """Extract metadata (e.g. PDB ID, comparative modeling templates) from a
       PDB file. This handles PDB headers added by the PDB database itself,
       comparative modeling packages such as MODELLER and Phyre2, and also
       some custom headers that can be used to indicate that a file has been
       locally modified in some way."""

    def parse_file(self, filename):
        """Extract metadata. See :meth:`Parser.parse_file` for details.

           :param str filename: the file to extract metadata from.
           :return: a dict with key `dataset` pointing to the PDB dataset;
                    'templates' pointing to a dict with keys the asym (chain)
                    IDs in the PDB file and values the list of comparative
                    model templates used to model that chain as
                    :class:`ihm.startmodel.Template` objects;
                    'entity_source' pointing to a dict with keys the asym IDs
                    and values :class:`ihm.source.Source` objects;
                    'software' pointing to a list of software used to generate
                    the file (as :class:`ihm.Software` objects);
                    'script' pointing to the script used to generate the
                    file, if any (as :class:`ihm.location.WorkflowFileLocation`
                    objects);
                    'metadata' a list of PDB metadata records.

           This parser looks at PDB headers. Standard PDB database headers are
           recognized, plus some added by common comparative modeling
           packages such as MODELLER and Phyre2, as well as some custom headers
           that can be used to denote that a PDB file is a locally-modified
           version of some other resource. Additional details will be extracted
           from other PDB headers if available, such as ``TITLE`` records.

           If the first line of the file starts with ``HEADER`` then the file
           is assumed to live in the PDB database. For example, the following
           will be interpreted as PDB entry 2HBJ::

               HEADER    HYDROLASE, GENE REGULATION              14-JUN-06   2HBJ

           If the first line starts with ``EXPDTA    DERIVED FROM`` then the
           file is assumed to derive from a given PDB ID or a comparative
           or integrative model available at a given DOI. ``TITLE`` records
           are expected to describe the nature of the transformation::

               EXPDTA    DERIVED FROM PDB:1YKH
               EXPDTA    DERIVED FROM COMPARATIVE MODEL, DOI:10.1093/nar/gkt704
               EXPDTA    DERIVED FROM INTEGRATIVE MODEL, DOI:10.1016/j.str.2017.01.006

           A first line starting with ``REMARK  99  Chain ID :`` is assumed to
           be a model generated by Phyre2. Template information can be added
           using Modeller-style headers, as below, if desired.

           A first line starting with ``EXPDTA    THEORETICAL MODEL, MODELLER``
           is assumed to be a model generated by Modeller. Headers generated
           by modern versions of Modeller are parsed to extract information
           about the comparative modeling script, plus the templates used and
           their alignment.
           Templates named ``1abcX`` or ``1abcX_N`` are assumed to be
           structures deposited in PDB (in this case, chain X in
           structure 1ABC).
           A custom ``TEMPLATE PATH`` header can be used to point to templates
           that are not deposited in the PDB database. For example, the model
           below is assumed to be constructed using templates from PDB codes
           3JRO and 3F3F, plus another template in ``my_custom_pdb_file.pdb``,
           and the given alignment::

               EXPDTA    THEORETICAL MODEL, MODELLER 9.18 2017/02/10 22:21:34
               REMARK   6 ALIGNMENT: modeller_model.ali
               REMARK   6 SCRIPT: model-default.py
               REMARK   6 TEMPLATE PATH custom1 ../inputs/my_custom_pdb_file.pdb
               REMARK   6 TEMPLATE: 3jroC 33:C - 424:C MODELS 33:A - 424:A AT 100.0%
               REMARK   6 TEMPLATE: 3f3fG 482:G - 551:G MODELS 429:A - 488:A AT 10.0%
               REMARK   6 TEMPLATE: custom1 9:A - 352:A MODELS 80:A - 414:A AT 32.0%

           A first line starting with ``TITLE     SWISS-MODEL SERVER``
           is assumed to be a model generated by SWISS-MODEL, and information
           about the template(s) is extracted from ``REMARK    3`` records.
        """  # noqa:  E501
        ret = {'templates': {}, 'software': [], 'metadata': [], 'script': None,
               'entity_source': {}}
        with open(filename) as fh:
            first_line = fh.readline()
            local_file = location.InputFileLocation(
                filename, details="Starting model structure")
            if first_line.startswith('HEADER'):
                self._parse_official_pdb(fh, first_line, ret)
            elif first_line.startswith('EXPDTA    DERIVED FROM PDB:'):
                self._parse_derived_from_pdb(fh, first_line, local_file,
                                             ret)
            elif first_line.startswith('EXPDTA    DERIVED FROM COMPARATIVE '
                                       'MODEL, DOI:'):
                self._parse_derived_from_comp_model(fh, first_line, local_file,
                                                    ret)
            elif first_line.startswith('EXPDTA    DERIVED FROM INTEGRATIVE '
                                       'MODEL, DOI:'):
                self._parse_derived_from_int_model(fh, first_line, local_file,
                                                   ret)
            elif first_line.startswith(
                    'EXPDTA    THEORETICAL MODEL, MODELLER'):
                self._parse_modeller_model(fh, first_line, local_file,
                                           filename, ret)
            elif first_line.startswith('REMARK  99  Chain ID :'):
                self._parse_phyre_model(fh, first_line, local_file,
                                        filename, ret)
            elif first_line.startswith('TITLE     SWISS-MODEL SERVER'):
                self._parse_swiss_model(fh, first_line, local_file,
                                        filename, ret)
            else:
                self._parse_unknown_model(fh, first_line, local_file,
                                          filename, ret)
        return ret

    def _parse_official_pdb(self, fh, first_line, ret):
        """Handle a file that's from the official PDB database."""
        version, details, metadata, entity_source \
            = self._parse_pdb_records(fh, first_line)
        loc = location.PDBLocation(first_line[62:66].strip(), version, details)
        ret['entity_source'] = entity_source
        ret['metadata'] = metadata
        ret['dataset'] = dataset.PDBDataset(loc, details=loc.details)

    def _parse_derived_from_pdb(self, fh, first_line, local_file, ret):
        # Model derived from a PDB structure; treat as a local experimental
        # model with the official PDB as a parent
        local_file.details = self._parse_details(fh)
        db_code = first_line[27:].strip()
        d = dataset.PDBDataset(local_file, details=local_file.details)
        d.parents.append(dataset.PDBDataset(location.PDBLocation(db_code)))
        ret['dataset'] = d

    def _parse_derived_from_comp_model(self, fh, first_line, local_file, ret):
        """Model derived from a comparative model; link back to the original
           model as a parent"""
        self._parse_derived_from_model(
            fh, first_line, local_file, ret, dataset.ComparativeModelDataset,
            'comparative')

    def _parse_derived_from_int_model(self, fh, first_line, local_file, ret):
        """Model derived from an integrative model; link back to the original
           model as a parent"""
        self._parse_derived_from_model(
            fh, first_line, local_file, ret, dataset.IntegrativeModelDataset,
            'integrative')

    def _parse_derived_from_model(self, fh, first_line, local_file, ret,
                                  dataset_class, model_type):
        local_file.details = self._parse_details(fh)
        d = dataset_class(local_file)
        repo = location.Repository(doi=first_line[46:].strip())
        # todo: better specify an unknown path
        orig_loc = location.InputFileLocation(
            repo=repo, path='.',
            details="Starting %s model structure" % model_type)
        d.parents.append(dataset_class(orig_loc))
        ret['dataset'] = d

    def _parse_modeller_model(self, fh, first_line, local_file, filename, ret):
        version, date = first_line[38:].rstrip('\r\n').split(' ', 1)
        s = ihm.Software(
            name='MODELLER', classification='comparative modeling',
            description='Comparative modeling by satisfaction '
                        'of spatial restraints, build ' + date,
            location='https://salilab.org/modeller/',
            version=version,
            citation=ihm.citations.modeller)
        ret['software'].append(s)
        self._handle_comparative_model(local_file, filename, ret)

    def _parse_phyre_model(self, fh, first_line, local_file, filename, ret):
        # Model generated by Phyre2
        s = ihm.Software(
            name='Phyre2', classification='protein homology modeling',
            description='Protein Homology/analogY Recognition '
                        'Engine V 2.0',
            version='2.0', location='http://www.sbg.bio.ic.ac.uk/~phyre2/',
            citation=ihm.citations.phyre2)
        ret['software'].append(s)
        self._handle_comparative_model(local_file, filename, ret)

    def _parse_swiss_model(self, fh, first_line, local_file, filename, ret):
        # Model generated by SWISS-MODEL
        meta = _get_swiss_model_metadata(filename)
        s = ihm.Software(
            name='SWISS-MODEL', classification='protein homology modeling',
            description='SWISS-MODEL: homology modelling of protein '
                        'structures and complexes, using %s engine'
                        % meta.get('info', {}).get('ENGIN', 'unknown'),
            version=meta.get('info', {}).get('VERSN', ihm.unknown),
            location='https://swissmodel.expasy.org/',
            citation=ihm.citations.swiss_model)
        ret['software'].append(s)
        comp_model_ds = dataset.ComparativeModelDataset(local_file)
        ret['dataset'] = comp_model_ds

        ret['templates'] = self._add_swiss_model_templates(
            local_file, meta, comp_model_ds, ret)

    def _add_swiss_model_templates(self, local_file, meta, comp_model_ds, ret):
        """Add template information extracted from SWISS-MODEL PDB metadata"""
        ret_templates = {}
        templates = [v for k, v in sorted(((k, v) for k, v in meta.items()
                                          if k.startswith('TEMPLATE')),
                                          key=operator.itemgetter(0))]
        for t in templates:
            loc = location.PDBLocation(t['PDBID'])
            d = dataset.PDBDataset(loc)
            # Make the comparative model dataset derive from the template's
            comp_model_ds.parents.append(d)
            for chain in t['MMCIF']:
                # todo: check we're using the right chain ID and that target
                # and template chain IDs really are always the same
                offset = int(t[chain, 'OFF'])
                tgt_seq, tgt_len = _parse_seq(t[chain, 'TRG'])
                tmpl_seq, tmpl_len = _parse_seq(t[chain, 'TPL'])
                tgt_rng, tmpl_rng = _get_aligned_region(tgt_seq, tmpl_seq)

                # apply offset
                tmpl_rng = (tmpl_rng[0] + offset, tmpl_rng[1] + offset)

                seq_id = float(t['SID'])
                seq_id = startmodel.SequenceIdentity(
                    float(t['SID']),
                    SequenceIdentityDenominator.NUM_ALIGNED_WITHOUT_GAPS)
                tmpl = startmodel.Template(
                    dataset=d, asym_id=chain, seq_id_range=tgt_rng,
                    template_seq_id_range=tmpl_rng, sequence_identity=seq_id,
                    alignment_file=local_file)
                ret_templates[chain] = [tmpl]
        return ret_templates

    def _parse_unknown_model(self, fh, first_line, local_file, filename, ret):
        # todo: revisit assumption that all unknown source PDBs are
        # comparative models
        self._handle_comparative_model(local_file, filename, ret)

    def _handle_comparative_model(self, local_file, pdbname, ret):
        d = dataset.ComparativeModelDataset(local_file)
        ret['dataset'] = d
        ret['templates'], ret['script'] \
            = self._get_templates_script(pdbname, d)

    def _get_templates_script(self, pdbname, target_dataset):
        template_path_map = {}
        alnfile = None
        script = None
        alnfilere = re.compile(r'REMARK   6 ALIGNMENT: (\S+)')
        scriptre = re.compile(r'REMARK   6 SCRIPT: (\S+)')
        tmppathre = re.compile(r'REMARK   6 TEMPLATE PATH (\S+) (\S+)')
        tmpre = re.compile(r'REMARK   6 TEMPLATE: '
                           r'(\S+) (\S+):(\S+) \- (\S+):\S+ '
                           r'MODELS (\S+):(\S+) \- (\S+):\S+ AT (\S+)%')
        template_info = []

        with open(pdbname) as fh:
            for line in fh:
                if line.startswith('ATOM'):  # Read only the header
                    break
                m = tmppathre.match(line)
                if m:
                    template_path_map[m.group(1)] = \
                        util._get_relative_path(pdbname, m.group(2))
                m = alnfilere.match(line)
                if m:
                    # Path to alignment is relative to that of the PDB file
                    fname = util._get_relative_path(pdbname, m.group(1))
                    alnfile = location.InputFileLocation(
                        fname,
                        details="Alignment for starting comparative model")
                m = scriptre.match(line)
                if m:
                    # Path to script is relative to that of the PDB file
                    fname = util._get_relative_path(pdbname, m.group(1))
                    script = location.WorkflowFileLocation(
                        fname, details="Script for starting comparative model")
                m = tmpre.match(line)
                if m:
                    template_info.append(m)

        templates = {}
        for t in template_info:
            chain, template = self._handle_template(t, template_path_map,
                                                    target_dataset, alnfile)
            if chain not in templates:
                templates[chain] = []
            templates[chain].append(template)
        # Sort templates by starting residue, then ending residue
        for chain in templates.keys():
            templates[chain] = sorted(templates[chain],
                                      key=operator.attrgetter('seq_id_range'))
        return templates, script

    def _handle_template(self, info, template_path_map, target_dataset,
                         alnfile):
        """Create a Template object from Modeller PDB header information."""
        template_code = info.group(1)
        template_seq_id_range = (int(info.group(2)), int(info.group(4)))
        template_asym_id = info.group(3)
        seq_id_range = (int(info.group(5)), int(info.group(7)))
        target_asym_id = info.group(6)
        sequence_identity = startmodel.SequenceIdentity(
            float(info.group(8)), SequenceIdentityDenominator.SHORTER_LENGTH)

        # Assume a code of 1abc, 1abc_N, 1abcX, or 1abcX_N refers
        # to a real PDB structure
        m = re.match(r'(\d[a-zA-Z0-9]{3})[a-zA-Z]?(_.*)?$', template_code)
        if m:
            template_db_code = m.group(1).upper()
            loc = location.PDBLocation(template_db_code)
        else:
            # Otherwise, look up the PDB file in TEMPLATE PATH remarks
            fname = template_path_map[template_code]
            loc = location.InputFileLocation(
                fname, details="Template for comparative modeling")
        d = dataset.PDBDataset(loc, details=loc.details)

        # Make the comparative model dataset derive from the template's
        target_dataset.parents.append(d)

        return (target_asym_id,
                startmodel.Template(
                    dataset=d, asym_id=template_asym_id,
                    seq_id_range=seq_id_range,
                    template_seq_id_range=template_seq_id_range,
                    sequence_identity=sequence_identity,
                    alignment_file=alnfile))

    def _parse_pdb_records(self, fh, first_line):
        """Extract information from an official PDB"""
        metadata = []
        details = ''
        compnd = ''
        source = ''
        for line in fh:
            if line.startswith('TITLE'):
                details += line[10:].rstrip()
            elif line.startswith('COMPND'):
                compnd += line[10:].rstrip()
            elif line.startswith('SOURCE'):
                source += line[10:].rstrip()
            elif line.startswith('HELIX'):
                metadata.append(startmodel.PDBHelix(line))
        return (first_line[50:59].strip(),
                details if details else None, metadata,
                self._make_entity_source(compnd, source))

    def _make_one_entity_source(self, compnd, source):
        """Make a single ihm.source.Source object"""
        def make_from_source(cls):
            return cls(scientific_name=source.get('ORGANISM_SCIENTIFIC'),
                       common_name=source.get('ORGANISM_COMMON'),
                       strain=source.get('STRAIN'),
                       ncbi_taxonomy_id=source.get('ORGANISM_TAXID'))
        if compnd.get('ENGINEERED', None) == 'YES':
            gene = make_from_source(ihm.source.Details)
            host = ihm.source.Details(
                scientific_name=source.get('EXPRESSION_SYSTEM'),
                common_name=source.get('EXPRESSION_SYSTEM_COMMON'),
                strain=source.get('EXPRESSION_SYSTEM_STRAIN'),
                ncbi_taxonomy_id=source.get('EXPRESSION_SYSTEM_TAXID'))
            return ihm.source.Manipulated(gene=gene, host=host)
        else:
            if source.get('SYNTHETIC', None) == 'YES':
                cls = ihm.source.Synthetic
            else:
                cls = ihm.source.Natural
            return make_from_source(cls)

    def _make_entity_source(self, compnd, source):
        """Make ihm.source.Source objects given PDB COMPND and SOURCE lines"""
        entity_source = {}
        # Convert each string into dict of mol_id vs keys
        compnd = self._parse_pdb_mol_id(compnd)
        source = self._parse_pdb_mol_id(source)
        for mol_id, c in compnd.items():
            if mol_id in source and 'CHAIN' in c:
                s = self._make_one_entity_source(c, source[mol_id])
                for chain in c['CHAIN'].split(','):
                    entity_source[chain.strip()] = s
        return entity_source

    def _parse_pdb_mol_id(self, txt):
        """Convert text COMPND or SOURCE records to a dict of mol_id vs keys"""
        d = {}
        mol_id = None
        for pair in txt.split(';'):
            spl = pair.split(':')
            if len(spl) == 2:
                key = spl[0].upper().strip()
                val = spl[1].upper().strip()
                if key == 'MOL_ID':
                    mol_id = d[val] = {}
                elif mol_id is not None:
                    mol_id[key] = val
        return d

    def _parse_details(self, fh):
        """Extract TITLE records from a PDB file"""
        details = ''
        for line in fh:
            if line.startswith('TITLE'):
                details += line[10:].rstrip()
            elif line.startswith('ATOM'):
                break
        return details
